"""
Configuration tools.
"""

import os


def read_pkginfo(dirname, infofile="pkginfo.py", versionfile="version.py",
                 versionattr="__version__", increment="0.1"):
    """
    Read python package information.

    Avoids importing as a module, since that might mean having other stuff
    installed before it needs to be (i.e., at setup stage).

    If the package information defines a version attribute, this is used
    as-is.  Otherwise, the version is extracted from the repository tags if
    possible (see hgtools) and written to a version file for use by the
    package.

    Args:
        dirname: Directory to read/write sources in.
        infofile: Package information file.
        versionfile: File to write version information to.
        versionattr: Info attribute storing the version.
        increment: Version auto-increment (see hgtools).

    Returns:
        Package information dict.
    """

    # Gather settings from the python file.
    info = Pkginfo()
    path = infopath = os.path.join(dirname, infofile)
    if os.path.exists(path):
        info.read(path)

    if versionattr not in info:
        # No version specified.
        version = None
        path = versionpath = os.path.join(dirname, versionfile)

        # Try getting it from repo.
        try:
            from hgtools.managers import MercurialManager
            mgr = MercurialManager()
            version = mgr.get_current_version(increment=increment)

            # Write it to version file, so that package can access it.
            with open(versionpath, "w") as fp:
                fp.write(version_format.format(versionattr, version))

            # Set the version attribute.
            info[versionattr] = version

        except (RuntimeError, ImportError):
            # Nope.  Try the version file.
            if os.path.exists(path):
                print("Not updating '{}' (no hgtools)".format(path))
                info.read(versionpath)

    # If still no version, give up.
    if versionattr not in info:
        raise RuntimeError(fail_message.format(infopath, versionpath))

    version = info[versionattr]
    print("Package version {} (from '{}')\n".format(version, path))
    return info


class Pkginfo(dict):
    "Helper class to get package information from Python sources."

    def read(self, filename):
        with open(filename) as fp:
            exec(fp.read(), self)

    def __getattr__(self, attr):
        return self[attr]


version_format = """# Autogenerated by setup -- do not edit!

{} = '{}'
"""

fail_message = """Can't determine package version

No version was found in either '{}' or '{}'.

If this is a Mercurial clone, you need to install the 'hgtools' module
before running setup.py so that the correct package version can be
determined from the repository.
"""



"""
A simple plugin system.

Inspiration:

* http://martyalchin.com/2008/jan/10/simple-plugin-framework
* https://pypi.python.org/pypi/pluginloader
"""

import os
import pkg_resources as pkg

from six import add_metaclass

from .logger import log


class PluginLoader(object):
    def __init__(self):
        #: List of paths and entrypoints to load.
        self.loading = []

        #: Set of paths and entrypoints already loaded.
        self.loaded = set()

    def add_path(self, path):
        "Add a path to the load list."
        self.loading.append(('path', path))

    def add_entrypoint(self, group):
        "Add an entrypoint group to the load list."
        self.loading.append(('entrypoint', group))

    def load(self):
        "Load all plugins."

        for tag, name in self.loading:
            if tag == 'path':
                self.load_directory(name)
            elif tag == 'entrypoint':   # pragma: no cover
                for entrypoint in pkg.iter_entry_points(name):
                    group = name + '.' + entrypoint.name
                    self.load_plugin(entrypoint, group)

    def load_plugin(self, entrypoint, group):
        if group not in self.loaded:
            try:
                log.info("loading plugin from entry point '%s'" % group)
                entrypoint.load()
            except Exception:
                log.exception('error loading %s (ignored)' % group)

            self.loaded.add(group)

    def load_file(self, filename, context=None):
        if filename not in self.loaded:
            log.info("loading plugins from file '%s'", filename)

            try:
                context = context or {}
                with open(filename) as fp:
                    exec(fp.read(), context)
            except Exception:
                log.exception('error loading %s (ignored)' % filename)

            self.loaded.add(filename)

    def load_directory(self, path, context=None):
        if path not in self.loaded and os.path.isdir(path):
            log.info("loading plugins from directory '%s'", path)

            for filename in os.listdir(path):
                if filename.endswith(".py"):
                    fullpath = os.path.join(path, filename)
                    self.load_file(fullpath, context)

            self.loaded.add(path)


class PluginMount(type):
    def __init__(cls, name, bases, attrs):
        if not hasattr(cls, 'plugins'):
            cls.plugins = {}
        elif hasattr(cls, 'name') and cls.name is not None:
            log.info("registering %s plugin: %s", cls.category, cls.name)
            cls.plugins[cls.category, cls.name] = cls


@add_metaclass(PluginMount)
class Plugin(object):
    #: Plugin name.  If None, it is not registered.
    name = None

    #: Plugin category.  This is set by direct subclasses which are
    #: intended to be subclassed further by third-party plugins.
    category = None

    #: One-line plugin description.  Appears in help text.
    description = "no description"

    #: Plugin version.  At the moment nothing is done with this; it's for
    #: information only.
    version = "0.1"

    #: Plugin author.  At the moment nothing is done with this; it's for
    #: information only.
    author = "unknown"

    #: Package name, if it is part of a package.  This is used to locate
    #: auxiliary files bundled in the package.
    package = None


def get_plugin(basecls, name):
    "Look up and return a plugin of a given type and name."

    for cls in get_plugins(basecls):
        if name == cls.name:
            return cls


def get_plugins(basecls):
    "Yield all plugins of a given type."

    for (name, category), cls in sorted(basecls.plugins.items()):
        if issubclass(cls, basecls):
            yield cls


# Create standard plugin loader.
loader = PluginLoader()



"""
Utility commands.
"""

import os


class CmdTools(object):
    def do_ipython(self, arg):
        "ipython -- enter embedded IPython interpreter"

        try:
            from IPython import embed
        except ImportError:
            self.error("IPython is not available")

        embed()

    def do_shell(self, arg):
        "shell -- run a system command"
        os.system(arg)



"""
Ditz command handler.
"""

from __future__ import print_function

import os
import sys
import six

from cmd import Cmd

from .command import CmdProject, CmdIssue, CmdInfo, CmdTools

from .database import DitzDB
from .settings import Settings
from .config import config as userconfig
from .config import ConfigSection
from .colorize import colorize
from .objects import Project, Config, find_config
from .files import write_config
from .plugin import Plugin, get_plugins
from .util import print_columns, default_name, default_email, DitzError

from six.moves import input


class DitzCmd(Cmd, CmdProject, CmdIssue, CmdInfo, CmdTools):
    #: The :class:`DitzDB` database.
    db = None

    #: Whether running interactively.
    interactive = False

    #: Last issue name mentioned.
    last_issuename = None

    #: Last release name mentioned.
    last_releasename = None

    def __init__(self, path=".", settings=None, interactive=False, **kw):
        Cmd.__init__(self)

        #: Command prompt.
        self.prompt = "Ditz: "

        #: Pager prompt.
        self.moreprompt = "-- More (Y/n/q) --"

        #: Global settings.
        self.settings = settings or Settings(**kw)

        #: Configuration variables.
        self.config = ConfigSection(None, 'command', userconfig)

        #: Pathname of the database.
        self.dbpath = os.path.realpath(path)

        self.interactive = interactive
        self.last_issuename = None
        self.last_releasename = None

        # In interactive mode, write status (and do interactive setup
        # if creating a new database).
        dirname, _ = find_config(self.dbpath,
                                 search=self.settings.searchparents)

        if not dirname:
            self.write("Setting up Ditz database in", path)
            self.write()
            self.setup()
        elif self.interactive:
            path = os.path.realpath(dirname)
            self.write("Reading Ditz database in", path)
            self.write()

        # Read the database.
        self.db = DitzDB.read(self.dbpath, settings=settings)

        # Register plugin commands.
        for cls in get_plugins(CommandSet):
            for name in dir(cls):
                if name.startswith(("do_", "help_")):
                    func = six.get_unbound_function(getattr(cls, name))
                    meth = six.create_bound_method(func, self)
                    setattr(DitzCmd, name, meth)

        # Set the list of known commands.
        self.commands = set()
        for name in self.get_names():
            if name.startswith("do_"):
                self.commands.add(name[3:])

        # Highlight the prompts if required.
        if self.interactive and self.settings.highlight:
            self.prompt = colorize(self.prompt.strip()) + " "
            self.moreprompt = colorize(self.moreprompt)

    def setup(self, oldconfig=None):
        """
        Set up a database interactively.
        """

        config = self.getconfig(oldconfig)
        if config:
            name = config.name
            email = config.email
            issuedir = config.issue_dir
        else:
            return False

        path = os.path.join(self.dbpath, issuedir, Project.filename)

        if os.path.exists(path):
            if write_config(config):
                self.write()
                self.write("Configuration written")
        else:
            default = os.path.basename(self.dbpath)
            project = self.getline("Project name", default)
            if not project:
                return False

            db = DitzDB(project, name, email, issuedir, self.dbpath,
                        settings=self.settings)
            db.write()

            self.write()
            self.write("Ditz database created in", self.dbpath)

        return True

    # Builtin command methods.

    def do_init(self, arg):
        "init -- Initialize the issue database for a new project"

        # This is a no-op, kept for Ditz compatibility, since
        # initialization is automatically done if no database exists.
        pass

    def do_reconfigure(self, arg):
        "reconfigure -- Rerun configuration script"

        self.setup(self.db.config)

    def do_help(self, arg):
        "help [command] -- Display help, optionally on a command"

        if arg:
            arg = self.findcmd(arg)

        if arg is not None:
            Cmd.do_help(self, arg)

    def do_quit(self, arg):
        "quit -- quit the interactive command loop"
        sys.exit()

    do_EOF = do_quit

    # Completion methods.

    def match_issue(self, text, line, beg, end):
        return startswith(self.db.issue_names, text)

    complete_add_reference = match_issue
    complete_assign = match_issue
    complete_close = match_issue
    complete_comment = match_issue
    complete_drop = match_issue
    complete_edit = match_issue
    complete_set_component = match_issue
    complete_show = match_issue
    complete_start = match_issue
    complete_stop = match_issue
    complete_unassign = match_issue

    def match_release(self, text, line, beg, end):
        return startswith(self.db.releases, text)

    complete_archive = match_release
    complete_changelog = match_release
    complete_release = match_release
    complete_status = match_release
    complete_todo = match_release

    def match_command(self, text):
        return startswith(self.commands, text)

    # Command hook methods.

    def onecmd(self, cmdline):
        # Get command and arguments.
        bits = cmdline.strip().split(None, 1)
        cmd = args = ""

        if len(bits) == 1:
            cmd = bits[0]
        elif len(bits) > 1:
            cmd, args = bits

        # Run it.
        retval = False

        try:
            cmd = self.config.name = self.findcmd(cmd)
            Cmd.onecmd(self, cmd + " " + args)
            retval = True
        except DitzError as msg:
            self.write("Error:", msg)
        except (KeyboardInterrupt, EOFError):
            self.write()
            self.write("Aborted")

        return None if self.interactive else retval

    def findcmd(self, cmd):
        if cmd:
            # Substitute aliases, if any.
            if userconfig.has_option('alias', cmd):
                cmd = userconfig.get('alias', cmd)

            # Replace hyphens in command, to support original Ditz-style
            # commands.
            cmd = cmd.replace('-', '_')

            # Do completion on it.
            if cmd not in self.commands:
                matches = self.match_command(cmd)
                if len(matches) == 1:
                    cmd = matches[0]
                elif len(matches) > 1:
                    self.error("ambiguous command: %s (%s)"
                               % (cmd, ", ".join(matches)))

        return cmd

    def emptyline(self):
        return self.onecmd("todo")

    def default(self, arg):
        msg = "unknown command: " + arg
        if self.interactive:
            msg += " (type 'help' for a list)"

        self.error(msg)

    # Input/output methods.

    def getarg(self, arg, idx):
        """
        Get a string argument.

        Args:
            arg (str): Raw command string.
            idx (int): Argument index (numbered from 1).

        Returns:
            Value (str), or None if out of range.
        """

        args = arg.split()
        if 0 < idx <= len(args):
            return args[idx - 1]

        return None

    def getint(self, arg, idx, default=None):
        """
        Get an integer argument.

        Args:
            arg (str): Raw command string.
            idx (int): Argument index (numbered from 1).
            default (int, optional): Default value.

        Returns:
            Value (int), or None.
        """

        value = self.getarg(arg, idx)
        if not value:
            return default

        try:
            return int(value)
        except ValueError:
            self.error("expected an integer, not", value)

    def getissue(self, arg, idx):
        """
        Get an issue by name or ID.  Return it and its assigned name.

        Args:
            arg (str): Raw command string.
            idx (int): Argument index (numbered from 1).

        Returns:
            Tuple(Issue, str).
        """

        name = self.getarg(arg, idx)
        if not name and self.last_issuename:
            name = self.last_issuename

        if not name:
            self.error("no issue specified")

        self.last_issuename = name
        issue = self.db.get_issue(name)

        if not issue:
            issues = [i for i in self.db.issues if i.id.startswith(name)]
            count = len(issues)
            if count == 1:
                issue = issues.pop()
            elif count > 1:
                self.error("%d issues match ID '%s'" % (count, name))

        if not issue:
            self.error("no issue with name or ID '%s'" % name)

        return issue, self.db.issue_name(issue)

    def getrelease(self, arg, idx, optional=False):
        """
        Get a release by name.

        Args:
            arg (str): Raw command string.
            idx (int): Argument index (numbered from 1).
            optional (bool): Whether to allow unspecified.

        Returns:
            Release name (str).
        """

        name = self.getarg(arg, idx)
        if not name and optional:
            return None

        if not name and self.last_releasename:
            name = self.last_releasename

        if not name:
            self.error("no release specified")

        self.last_releasename = name
        if name not in self.db.releases:
            self.error("no release with name '%s'" % name)

        return name

    def getline(self, prompt="> ", default="", allowempty=True):
        """
        Get a single line of input.

        Args:
            prompt (str): Prompt string.
            default (str): Default value.
            allowempty (bool): Whether value is optional.

        Returns:
            Reply (str).
        """

        if default:
            prompt = "%s (default '%s'): " % (prompt, default)

        while True:
            if self.use_rawinput:
                reply = input(prompt)
            else:
                self.stdout.write(prompt)
                self.stdout.flush()
                reply = self.stdin.readline()

            if not len(reply.strip()):
                reply = default

            reply = reply.rstrip('\r\n')
            if reply or allowempty:
                return reply

    def gettext(self, title=None, prompt="> ", endchar='.'):
        """
        Get multiline text.

        Args:
            title (str): Initial title text.
            prompt (str): Prompt for each line of input.
            endchar (str): Character terminating input.

        Returns:
            Text (str).
        """

        if title:
            self.write(title + " (ctrl-c to abort, %s to finish)" % endchar)

        lines = []
        while True:
            line = self.getline(prompt)
            if line == endchar:
                return "\n".join(lines)

            lines.append(line)

    def getchoice(self, thing, choices):
        """
        Get a choice of several things.

        Args:
            thing (str): Description of thing being chosen.
            choices (list): List of available options.

        Returns:
            Choice made (str).
        """

        items = []
        for num, entry in enumerate(choices, 1):
            item = "%3d) %s" % (num, entry)
            items.append(item)

        print_columns(items)
        prompt = "Choose a %s (1--%d): " % (thing, len(choices))

        while True:
            reply = self.getline(prompt)

            try:
                return choices[int(reply) - 1]
            except (ValueError, IndexError):
                pass

    def getcomment(self):
        """
        Get a comment string.

        Returns:
            Comment text (str).
        """

        if self.settings.nocomment:
            return ""
        elif not self.interactive and self.settings.comment:
            return self.settings.comment
        else:
            return self.gettext("Comments")

    def getyesno(self, question, default=False):
        """
        Get the answer to a yes/no question.

        Args:
            question (str): The question.
            default (bool): Default if no reply given.

        Returns:
            Reply (bool).
        """

        prompt = question + " [%s] " % ("yes" if default else "no")

        while True:
            reply = self.getline(prompt)
            if not reply:
                return default
            if reply[0] in "yY":
                return True
            elif reply[0] in "nN":
                return False

    def getconfig(self, oldconfig=None):
        """
        Prompt for and return database configuration info.
        """

        default = oldconfig.name if oldconfig else default_name()
        name = self.getline("Your name", default)

        default = oldconfig.email if oldconfig else default_email()
        email = self.getline("Your email", default)

        issuedirs = userconfig.get('config', 'issuedirs').split()
        issuedir = issuedirs[0] if issuedirs else ".ditz-issues"
        default = oldconfig.issue_dir if oldconfig else issuedir
        issuedir = self.getline("Issue directory", default)

        return Config(name, email, issuedir)

    def write(self, *args):
        """
        Write the given list of arguments to stdout.

        Also performs syntax highlighting and paging.
        """

        if args:
            args = [six.text_type(s) for s in args]
            items = " ".join(args).split("\n")
            lines = self.settings.termlines

            for num, line in enumerate(items, 1):
                if not six.PY3:
                    line = line.encode('utf-8')

                if self.settings.highlight:
                    line = colorize(line)

                print(line)
                if lines and num % (lines - 1) == 0:
                    reply = input(self.moreprompt)
                    if reply and reply[0] in "nqNQ":
                        break
        else:
            print()

    def error(self, *args):
        """
        Signal an error.
        """

        raise DitzError(" ".join(args))

    # Miscellaneous other stuff.

    def unimplemented(self):
        """
        Write a 'not implemented yet' message.
        """

        cmd = self.lastcmd.split()[0]
        self.error("'%s' is not implemented yet" % cmd)


class CommandSet(Plugin):
    """
    Base class for plugin commands.
    """

    category = "command"

    #: Name of the command set.
    name = None

    #: One-line description (currently unused).
    description = "undocumented"


def startswith(items, text):
    return [x for x in items if x.startswith(text)]



"""
Issue, release and status flags.
"""

# Issue type.

BUGFIX = ":bugfix"
FEATURE = ":feature"
TASK = ":task"

TYPE = {BUGFIX: "bugfix", FEATURE: "feature", TASK: "task"}
TYPE_PLURAL = {BUGFIX: "bugfixes", FEATURE: "features", TASK: "tasks"}

# Issue status.

UNSTARTED = ":unstarted"
PAUSED = ":paused"
IN_PROGRESS = ":in_progress"
CLOSED = ":closed"

STATUS = {UNSTARTED: "unstarted", PAUSED: "paused",
          IN_PROGRESS: "in progress", CLOSED: "closed"}

FLAGS = {UNSTARTED: "_", PAUSED: "=", IN_PROGRESS: ">", CLOSED: "x"}
SORT = {UNSTARTED: 1, PAUSED: 2, IN_PROGRESS: 3, CLOSED: 0}

# Issue disposition.

FIXED = ":fixed"
WONTFIX = ":wontfix"
REORG = ":reorg"

DISPOSITION = {FIXED: "fixed", WONTFIX: "won't fix", REORG: "reorganized"}

# Release status.

RELEASED = ":released"
UNRELEASED = ":unreleased"

RELSTATUS = {RELEASED: "released", UNRELEASED: "unreleased"}



"""
Logging facilities.
"""

import logging
from .pkginfo import __title__

#: The logging object.
log = logging.getLogger(__title__)

# Add a null handler.
log.addHandler(logging.NullHandler())


def init_logging(verbose=False, formatstring="# %(message)s"):
    # Add a stderr handler.
    handler = logging.StreamHandler()
    log.addHandler(handler)

    fmt = logging.Formatter(formatstring)
    handler.setFormatter(fmt)

    # Set log level.
    log.setLevel(logging.INFO if verbose else logging.WARNING)



"""
Mixins for builtin commands.
"""

from .info import CmdInfo               # noqa
from .issue import CmdIssue             # noqa
from .project import CmdProject         # noqa
from .tools import CmdTools             # noqa



"""
Common exporter stuff.
"""

import os
import shutil

from collections import defaultdict

import pkg_resources as pkg
from jinja2 import Environment, ChoiceLoader, PackageLoader, FileSystemLoader

from .plugin import Plugin, get_plugin, get_plugins
from .config import config, config_path, ConfigSection
from .util import make_directory, timespan
from .files import write_file
from .logger import log
from .pkginfo import __url__
from .version import __version__

from . import flags

#: Mapping of exporter names to their classes.
exporters = None

#: Mapping of suffixes to archive formats.
archive_suffix = {".tar": "tar",
                  ".tar.gz": "gztar",
                  ".tgz": "gztar",
                  ".tar.bz2": "bztar",
                  ".zip": "zip"}


def get_exporter(name):
    "Get an exporter class given its name."

    return get_plugin(Exporter, name)


def get_exporters():
    "Yield all available exporters and their descriptions."

    for cls in get_plugins(Exporter):
        yield cls.name, cls.description


class Exporter(Plugin):
    """
    Base class for database exporters.

    Args:
        database (DitzDB): Database object.
    """

    category = "exporter"

    #: Name of the exporter.  This is the argument given to the ``export``
    #: command.
    name = None

    #: One-line description.  This gets printed when you list the
    #: available export formats.
    description = "undocumented"

    #: Package name.  If exporter is being installed via setuptools' plugin
    #: system, this should be set so that static and template files can be
    #: found in the package.
    package = None

    #: Exporter file suffix.  This is usually the same as the exporter name
    #: string.
    suffix = None

    #: If the exporter uses static files, this is the subdirectory of
    #: ``static`` to find them.  Usually the same as the ``name``
    #: attribute.
    static_dir = None

    #: If the exporter uses templates, this is the subdirectory of
    #: ``templates`` to find them.  Usually the same as the ``name``
    #: attribute.
    template_dir = None

    def __init__(self, database):
        #: The issue database being exported.
        self.db = database

        #: Configuration variables.
        self.config = ConfigSection(self.name, 'export', config)

        #: Mapping of template names to loaded templates.
        self.templates = {}

        #: Template environment.
        self.env = None

        # Build search paths to static and template files.
        self.paths = defaultdict(list)

        for name, dirname in (("static", self.static_dir),
                              ("templates", self.template_dir)):
            if dirname is not None:
                for dirpath in (self.db.issuedir, config_path()):
                    path = os.path.join(dirpath, name, dirname)
                    self.paths[name].append(path)

        # Set up template environment if required.
        if self.template_dir is not None:
            paths = self.paths['templates']
            templates = os.path.join('templates', self.template_dir)
            package = self.package or __name__

            loaders = [FileSystemLoader(paths),
                       PackageLoader(package, templates)]

            self.env = Environment(loader=ChoiceLoader(loaders),
                                   trim_blocks=True, lstrip_blocks=True)

        # Add common filters.
        @self.add_filter
        def issues(item):
            attr = item.__class__.__name__.lower()
            return [issue for issue in self.db.issues
                    if getattr(issue, attr) == item.name]

        rmap = {rel.name: rel for rel in self.db.project.releases}
        cmap = {comp.name: comp for comp in self.db.project.components}
        relmap = self.db.relation_mapping()

        @self.add_filter
        def release(issue):
            return rmap.get(issue.release, None)

        @self.add_filter
        def component(issue):
            return cmap[issue.component]

        @self.add_filter
        def issuetype(issue):
            return flags.TYPE[issue.type]

        @self.add_filter
        def inprogress(issue):
            time = issue.progresstime
            return timespan(time) if time > 0 else ""

        @self.add_filter
        def related(issue):
            return list(sorted(relmap[issue]))

        @self.add_filter
        def dateformat(value, format="%Y-%m-%d"):
            return value.strftime(format)

        @self.add_filter
        def timeformat(value, format="%Y-%m-%d %H:%M"):
            return value.strftime(format)

        self.setup()

    def export(self, dirname):
        """
        Export the issue database to the given directory.

        Args:
            dirname (str): Directory to write files into.
        """

        # If output directory looks like the name of an archive, deduce
        # format and adjust the name.
        archive = None
        for suffix in archive_suffix:
            if dirname.endswith(suffix):
                dirname = dirname.replace(suffix, "")
                archive = archive_suffix[suffix]

        # Create output directory.
        make_directory(dirname, archive)

        # Preload all the templates if required.
        if self.template_dir is not None:
            for name in self.env.list_templates():
                self.templates[name] = template = self.env.get_template(name)
                log.info("loaded template from %s" % template.filename)

        # Write the exported files.
        self.write(dirname)

        # Copy static files, if any.  First those in the reversed search
        # path, so that earlier paths will override later ones.  Then copy
        # any files from the module that don't yet exist.
        if self.static_dir is not None:
            for path in reversed(self.paths['static']):
                if os.path.exists(path):
                    for name in os.listdir(path):   # pragma: no cover
                        filename = os.path.join(path, name)
                        shutil.copy(filename, dirname)
                        log.info("copied %s to %s" % (filename, dirname))

            static = os.path.join('static', self.static_dir)
            package = self.package or __name__

            if pkg.resource_exists(package, static):
                for name in pkg.resource_listdir(package, static):
                    src = os.path.join(static, name)
                    dst = os.path.join(dirname, name)
                    if not os.path.exists(dst):
                        filename = pkg.resource_filename(package, src)
                        shutil.copy(filename, dirname)
                        log.info("copied %s to %s" % (filename, dirname))

        # Create archive and remove directory if required.
        if archive:
            shutil.make_archive(dirname, archive, root_dir=".",
                                base_dir=dirname, logger=log, verbose=True)
            shutil.rmtree(dirname)

    def setup(self):
        """
        Do exporter-specific setup.

        By default, this does nothing.
        """

    def add_filter(self, func, name=None):
        """
        Add a custom Jinja filter.

        This returns the original function, so the method can be used as a
        decorator.

        Args:
            func (callable): Filter function.
            name (str, optional): Name to use in templates (same as
                function name, if None).

        Returns:
            func: original function
        """

        if self.env:
            self.env.filters[name or func.__name__] = func

        return func

    def write(self, dirname):
        """
        Write exported files to the given directory.

        This method must be overridden.  If using templates, it should call
        the :func:`render` method.

        Args:
            dirname (str): Directory to write files into.
        """

        raise NotImplementedError

    def render(self, dirname, templatefile, targetfile=None, **kw):
        """
        Render a single file from a template.

        Args:
            dirname (str): Directory to write file into.
            templatefile (str): Jinja template to use.
            targetfile (str, optional): Filename to write (same as
                 *templatefile*, if None).
            kw (dict): Template parameters.
        """

        # Get template.
        template = self.templates[templatefile]

        # Render template.
        kw.update(version=__version__, url=__url__)
        text = template.render(**kw)

        # Write it to file.
        path = os.path.join(dirname, targetfile or templatefile)
        write_file(path, text)
        log.info("wrote '%s'", path)

    def export_filename(self, item):
        """
        Return a unique export filename for a Ditz item.

        The filename is based on the item's ID (if it's an issue) or its
        name.  Each filename has the exporter suffix appended to it.

        Args:
            item (DitzObject): Ditz item.

        Return:
            Filename string.
        """

        clsname = item.__class__.__name__.lower()
        name = getattr(item, "id", None) or getattr(item, "name")
        return "%s-%s.%s" % (clsname, name, self.suffix)



"""
Database export tests.
"""

from ditz.database import DitzDB, DitzError
from nose.tools import raises


def test_html(name="example"):
    "Test exporting HTML"

    db = DitzDB.read(name)
    db.export('html', name + "-html")


def test_zip(name="example"):
    "Test exporting HTML ZIP file."

    db = DitzDB.read(name)
    db.export('html', name + "-issues.zip")


@raises(DitzError)
def test_zip_fail(name="example"):
    "Test exporting HTML ZIP file (not clobbering existing directory)."

    db = DitzDB.read(name)
    db.export('html', name + "-html.zip")



"""
Exporting to HTML.
"""

import re

from ditz.exporter import Exporter
from ditz.util import html_markup_function
from ditz import flags


class HTMLExporter(Exporter):
    """
    A HTML database exporter.
    """

    name = suffix = static_dir = template_dir = 'html'
    description = 'export to static HTML pages'

    markupfunc = None

    def setup(self):
        "Setup function.  Defines some HTML-specific filters."

        # Define HTML-specific filter functions.
        @self.add_filter
        def progressmeter(value, size=50):
            done = int(value * size)
            undone = max(0, size - done)
            return ("<span class='progress-meter'>" +
                    "<span class='progress-meter-done'>" +
                    ("&nbsp;" * done) +
                    "</span><span class='progress-meter-undone'>" +
                    ("&nbsp;" * undone) +
                    "</span></span>")

        @self.add_filter
        def imagetag(issue, cls="inline-status-image"):
            if issue.status == flags.UNSTARTED:
                return ""

            text = flags.STATUS[issue.status]

            if issue.status == flags.CLOSED:
                text = flags.DISPOSITION[issue.disposition]

                if issue.disposition == flags.FIXED:
                    image = "green-check.png"
                elif issue.disposition == flags.WONTFIX:
                    image = "red-check.png"
                elif issue.disposition == flags.REORG:
                    image = "blue-check.png"
            elif issue.status == flags.IN_PROGRESS:
                image = "green-bar.png"
            elif issue.status == flags.PAUSED:
                image = "yellow-bar.png"

            return '<img class="%s" alt="%s" title="%s" src="%s">' % \
                   (cls, text, text, image)

        @self.add_filter
        def sortkey(issue):
            text = issue.status
            if issue.closed:
                text += " " + issue.disposition
            return text

        @self.add_filter
        def link(item, text=None, image=False, cls=None):
            if not item:
                return ""

            if not text:
                text = item.name

            filename = self.export_filename(item)
            linktext = '<a href="%s">%s</a>' % (filename, text)

            if image:
                linktext += ' ' + imagetag(item)

            if cls:
                linktext = '<span class="%s">%s</span>' % (cls, linktext)

            return linktext

        idmap = {}
        for issue in self.db.issues:
            idmap[issue.id] = link(issue, image=True, cls="inline-issue-link")

        @self.add_filter
        def addlinks(text):
            return self.db.convert_to_name(text, idmap)

        obscure = re.compile(r'(.+)@[\w.]+(.*)')

        @self.add_filter
        def email(text):
            m = obscure.search(text)
            if m:
                return m.group(1) + '@...' + m.group(2)
            else:
                return text

        @self.add_filter
        def markup(text):
            if self.markupfunc is None:
                name = self.config.get('markup')
                self.markupfunc = html_markup_function(name)

            return self.markupfunc(text)

        # Markup style used in comment and description text.
        self.config.add('markup', 'markdown')

        # Number of recent-activity events shown on the index page.
        self.config.add('index_events', 10)

        # Number of recent-activity events shown on the release page.
        self.config.add('release_events', 10)

    def write(self, dirname):
        "Write HTML pages according to templates."

        project = self.db.project
        showcomp = len(self.db.components) >= 2

        # Write index page.
        events = sorted(self.db.issue_events,
                        key=lambda x: x[0], reverse=True)

        count = self.config.getint('index_events')
        self.render(dirname, 'index.html', project=project,
                    issues=self.db.issues, activity=events[:count])

        # Write a page for each release.
        count = self.config.getint('release_events')

        for rel in project.releases:
            issues = [i for i in self.db.issues if i.release == rel.name]
            issues = sorted(issues, key=lambda x: x.creation_time)
            issues = list(reversed(issues))

            relevents = [e for e in events if e[4].release == rel.name]

            filename = self.export_filename(rel)
            self.render(dirname, "release.html", targetfile=filename,
                        project=project, release=rel, issues=issues,
                        activity=relevents[:count], show_components=showcomp)

        # Write a page for each component.
        for comp in project.components:
            issues = [i for i in self.db.issues if i.component == comp.name]
            issues = sorted(issues, key=lambda x: x.creation_time)
            issues = list(reversed(issues))

            filename = self.export_filename(comp)
            self.render(dirname, "component.html", targetfile=filename,
                        project=project, component=comp, issues=issues)

        # Write a page for each issue.
        for issue in self.db.issues:
            filename = self.export_filename(issue)
            self.render(dirname, "issue.html", targetfile=filename,
                        project=project, issue=issue)

        # Write a page for unassigned issues.
        issues = [i for i in self.db.issues if not i.release]
        issues = sorted(issues, key=lambda x: x.creation_time)
        issues = list(reversed(issues))

        self.render(dirname, 'unassigned.html', project=project,
                    issues=issues, show_components=showcomp)



"""
Configuration file.
"""

import os
import re
import warnings

import pkg_resources as pkg
from six.moves import configparser as conf

from .pkginfo import __title__
from .logger import log


class DitzConfig(conf.ConfigParser):
    def __init__(self):
        conf.ConfigParser.__init__(self)
        defaults = pkg.resource_filename(__name__, "config.cfg")

        with open(defaults) as fp:
            if hasattr(self, "read_file"):
                self.read_file(fp)
            else:
                self.readfp(fp)

    def write_file(self, path):
        """Write config data to file."""

        with open(path, "w") as fp:
            fp.write("# %s configuration file.\n\n" % __title__)
            self.write(fp)


class Config(object):
    def __init__(self):
        self.path = get_userconfig()
        self.parser = None

    def set_file(self, path):
        self.path = path

    def set_option(self, setting):
        m = re.match(r'(\w+)\.(\w+)=(.*)', setting)
        if m:
            section, option, value = m.groups()
        else:
            raise ValueError("'%s': expected 'section.option=value'"
                             % setting)

        try:
            self.set(section, option, value)
        except conf.NoSectionError:
            raise ValueError("'%s': no such config section" % section)

    def __getattr__(self, attr):
        if not self.parser:
            self.parser = DitzConfig()
            log.info("reading %s" % self.path)
            self.parser.read(self.path)

        return getattr(self.parser, attr)


class ConfigSection(object):
    """
    Wrapper to the config file which looks up a particular section.
    """

    def __init__(self, name, section, config):
        self.name = name
        self.config = config
        self.section = section

    def add(self, name, default):
        """
        Add a named configuration value.
        """

        name = self.option(name)
        if not self.config.has_option(self.section, name):
            self.config.set(self.section, name, str(default))

    def get(self, name):
        """
        Get the named configuration value as a string.
        """

        return self.config.get(self.section, self.option(name))

    def getint(self, name):
        """
        Get the named configuration value as an integer.
        """

        return self.config.getint(self.section, self.option(name))

    def getfloat(self, name):
        """
        Get the named configuration value as a float.
        """

        return self.config.getfloat(self.section, self.option(name))

    def getboolean(self, name):
        """
        Get the named configuration value as a boolean.
        """

        return self.config.getboolean(self.section, self.option(name))

    def option(self, name):
        return self.name + "_" + name


def config_path(filename=None):
    "Return a pathname in the user's config directory."

    parts = [os.path.expanduser("~"), ".ditz"]

    if filename:
        parts.append(filename)

    return os.path.join(*parts)


def get_userconfig():
    "Return pathname of user config file."

    # Warn about old ~/.ditzrc if it exists.
    homedir = os.path.expanduser("~")
    path = os.path.join(homedir, ".ditzrc")
    if os.path.exists(path):
        warn("Move deprecated ~/.ditzrc file to ~/.ditz/ditz.cfg")
        return path

    # Try standard config file.
    return config_path("ditz.cfg")


def warn(msg):
    "Warn about obsolete configuration."
    warnings.warn(msg, UserWarning, stacklevel=2)


# Ditz user config settings.
config = Config()



"""
Default global settings.
"""


class Settings(object):
    def __init__(self, **kw):
        #: Whether to save files when changes are made.
        self.autosave = False

        #: Whether to use issue cache.
        self.usecache = False

        #: Issue cache filename.
        self.cachefile = ".ditz-cache"

        #: Whether to load external plugins.
        self.externalplugins = True

        #: Whether to do syntax highlighting.
        self.highlight = False

        #: Whether to skip asking for comments.
        self.nocomment = False

        #: Comment string to use in non-interactive mode.
        self.comment = None

        #: Whether to search in parent directories for database.
        self.searchparents = False

        #: Whether to enable version control.
        self.versioncontrol = False

        #: No. of terminal lines.  Zero turns off paging.
        self.termlines = 0

        # Set keyword values.
        for attr, val in kw.items():
            if hasattr(self, attr):
                setattr(self, attr, val)
            else:
                raise ValueError("no setting called '%s'" % attr)



"""
Project commands.
"""


class CmdProject(object):
    def do_add_component(self, arg):
        "add-component [name] -- Add a component"

        name = self.getarg(arg, 1) or self.getline("Name: ", allowempty=False)

        self.db.add_component(name)
        self.write("Added component", name)

    def do_add_release(self, arg):
        "add-release [name] -- Add a release"

        name = self.getarg(arg, 1) or self.getline("Name: ", allowempty=False)

        self.write("Adding release", name)
        comment = self.getcomment()

        self.db.add_release(name, comment=comment)
        self.write("Added release", name)

    def do_release(self, arg):
        "release <unreleased_release> -- Release a release"

        name = self.getrelease(arg, 1)
        comment = self.getcomment()

        self.db.release_release(name, comment)
        self.write("Released", name)

    def do_export(self, arg):
        "export <format> [path] -- Export issue database"

        fmt = self.getarg(arg, 1)
        if not fmt:
            self.error("no export format specified")

        path = self.getarg(arg, 2) or fmt
        self.db.export(fmt, path)
        self.write("Exported issues to '%s'" % path)

    def help_export(self):
        from ditz.exporter import get_exporters

        self.write(self.do_export.__doc__)
        self.write()
        self.write("available formats:")

        for name, desc in get_exporters():
            self.write("    %s -- %s" % (name, desc))

    def do_archive(self, arg):
        "archive <release> [dir] -- Archive a release"

        name = self.getrelease(arg, 1)
        path = self.getarg(arg, 2) or "ditz-archive-%s" % name

        self.db.archive_release(name, path)
        self.write("Archived to", path)

    def do_validate(self, arg):
        "validate -- Fix or report problems with issue database"

        changes = self.db.validate()

        if changes:
            for idx, msglist in sorted(changes.items()):
                for msg in msglist:
                    self.write("issue %s: %s" % (idx, msg))
        else:
            self.write("No problems found")



"""
Demo plugin.
"""

from ditz.commands import CommandSet

class AdventCommands(CommandSet):
    name = 'advent'
    description = 'adventuring commands'

    def do_xyzzy(self, arg):
        self.write("Nothing happens.")

    def help_xyzzy(self):
        self.write("xyzzy -- a secret magic word")

    def do_plugh(self, arg):
        "plugh -- another secret magic word"
        self.write("Nothing happens.")



"""
Text editor.
"""

from __future__ import print_function

import os
import yaml
import tempfile

from . import util
from .files import read_file, read_yaml_file
from .util import DitzError


class DitzEditor(object):
    """
    An editor for Ditz objects.
    """

    prefix = 'tmp-'
    suffix = '.yaml'
    textmode = False
    filetag = "<input>"

    def __init__(self, path):
        #: Original text.
        self.original = read_file(path)

        #: The text after editing.
        self.text = self.original

        #: Error message from last edit, or None.
        self.error = None

    def edit(self):
        "Edit the text and return it."

        # Get editor program.
        program = util.editor()
        if not program:
            raise DitzError("no text editor is configured")

        # Write text to a temp file.
        fp, filename = tempfile.mkstemp(prefix=self.prefix,
                                        suffix=self.suffix,
                                        text=self.textmode)
        os.write(fp, self.text.encode('utf-8'))
        os.close(fp)

        try:
            # Run the editor.
            util.run_editor(program, filename)

            # Get the edited text.
            self.text = read_file(filename)

            # Check for validity.
            err = self.validate(filename)
            if err:
                self.error = err.replace(filename, self.filetag)
            else:
                self.error = None

        finally:
            # Clean up.
            os.remove(filename)

        return self.text

    def validate(self, path):
        "Return validation error text or None."

        try:
            obj = read_yaml_file(path)
            obj.validate()
        except yaml.error.YAMLError as err:
            return str(err)
        except DitzError as err:
            return str(err)

        return None

    @property
    def modified(self):
        "Whether text has been modified."
        return self.text != self.original


if __name__ == "__main__":
    path = "../bugs/issue-3d1596d37cbd170f512eb939b97f5f22a3834c79.yaml"
    from ditz.objects import DitzObject   # noqa

    editor = DitzEditor(path)
    print(editor.edit())

    print("Modified:", editor.modified)
    print("Error:", editor.error)



# -*- coding: utf-8 -*-

"""
Database tests.
"""

import shutil
from filecmp import dircmp
from nose.tools import raises

from ditz.database import DitzDB, DitzError
from ditz.display import *
from ditz.flags import *

desc = u"""
A description cöntaining Unicode
"""


def test_create(name="database"):
    "Test database creation"

    try:
        shutil.rmtree(name)
    except OSError:
        pass

    db = DitzDB("MyProject", path=name, usecache=True, autosave=True)
    print(db)

    db.add_component("default")
    db.add_release('0.1')
    db.add_release('0.2')

    issue = db.add_issue("A title with Unicodé", desc)
    db.add_reference(issue, "file.txt")

    issue = db.add_issue("Critical fix", desc, comment="Really urgent")

    db.set_release(issue, '0.1')
    db.set_release(issue, '0.2')
    db.set_release(issue, None)

    db.add_comment(issue, "Will you make up yöur mind?")

    db.set_status(issue, IN_PROGRESS, comment="Let's do it!")
    db.set_status(issue, PAUSED, comment="Er.. let's think about it")
    db.set_status(issue, CLOSED, WONTFIX, comment="Nah, let's not bother")

    for issue in db:
        print(issue)

    for issuename in db.issue_names:
        print(issuename)

    db.write()

    issue = db.add_issue("Another issue")
    db.write()

    db.drop_issue(issue)
    db.write()

    newdb = DitzDB.read(name, usecache=True)
    assert len(newdb.issues) == 2

    newdb = DitzDB.read(name, usecache=False)
    assert len(newdb.issues) == 2


@raises(DitzError)
def test_nosuchcomp():
    "Test nonexistent component"

    db = DitzDB("MyProject")
    db.add_issue("Issue 1", component='fred')


def test_release():
    "Test valid release"

    db = DitzDB("MyProject")
    db.add_release("0.1")
    db.add_issue("Issue 1", release='0.1', status=CLOSED)
    db.release_release("0.1")


@raises(DitzError)
def test_norelease():
    "Test nonexistent release"

    db = DitzDB("MyProject")
    db.add_component("default")
    db.add_issue("Issue 1", release='1.0')


@raises(DitzError)
def test_duprelease():
    "Test duplicate releases"

    db = DitzDB("MyProject")
    db.add_release("0.2")
    db.add_release("0.2")


@raises(DitzError)
def test_invalidrelease_1():
    "Test invalid release (no issues)"

    db = DitzDB("MyProject")
    db.add_release("0.1")
    db.release_release("0.1")


@raises(DitzError)
def test_invalidrelease_2():
    "Test invalid release (issue not closed)"

    db = DitzDB("MyProject")
    db.add_release("0.1")
    db.add_issue("Issue 1", release='0.1')
    db.release_release("0.1")


@raises(DitzError)
def test_invalidrelease_3():
    "Test invalid release (no such release)"

    db = DitzDB("MyProject")
    db.release_release("0.1")


@raises(DitzError)
def test_invalidrelease_4():
    "Test invalid release (already released)"

    db = DitzDB("MyProject")
    db.add_release("0.1")
    db.add_issue("Issue 1", release='0.1', status=CLOSED)
    db.release_release("0.1")
    db.release_release("0.1")


@raises(DitzError)
def test_dupcomponent():
    "Test duplicate components"

    db = DitzDB("MyProject")
    db.add_component("doc")
    db.add_component("doc")


def test_show_todo():
    "Test show todo"

    db = DitzDB("MyProject")
    db.add_release("0.1")
    show_todo(db)
    show_todo(db, "0.1")
    show_todo(db, None, True)


def test_show_changelog():
    "Test show changelog"

    db = DitzDB("MyProject")
    db.add_release("0.1")
    db.add_issue("Issue 1", release='0.1', status=CLOSED)
    db.add_issue("Issue 2", release='0.1', status=CLOSED, type=FEATURE)
    show_changelog(db, "0.1")
    db.release_release("0.1")
    show_changelog(db, "0.1")


def test_archive():
    "Test archiving of releases"

    archive = "example-archive"

    try:
        shutil.rmtree(archive)
    except OSError:
        pass

    db = DitzDB("MyProject")
    db.add_release("0.1")
    db.add_issue("Issue 1", release='0.1', status=CLOSED)
    db.release_release("0.1")
    db.archive_release("0.1", archive)


def test_copy(name="example"):
    "Test database copying and roundtripping"

    db = DitzDB.read(name)
    db.write(name + "-copy")

    def check_files(dcmp):
        for name in dcmp.diff_files:
            raise Exception("file '%s' differs" % name)

        for sub_dcmp in list(dcmp.subdirs.values()):
            check_files(sub_dcmp)

    dcmp = dircmp(name, name + "-copy")
    check_files(dcmp)


def test_convert():
    "Test conversion between names and IDs in strings"

    db = DitzDB("MyProject")
    issue = db.add_issue("A bug")

    text = db.convert_to_id("This is issue ID myproject-1")
    assert text == "This is issue ID {issue %s}" % issue.id

    text = db.convert_to_name("This is issue name {issue %s}" % issue.id)
    assert text == "This is issue name myproject-1"

    name = db.issue_name(issue)
    assert name == "myproject-1"

    otherissue = db.get_issue(name)
    assert otherissue == issue



"""
Package information.
"""

__title__ = "pyditz"
__desc__ = "Python implementation of Ditz (http://rubygems.org/gems/ditz)."
__author__ = 'Glenn Hutchings'
__email__ = 'zondo42@gmail.com'
__license__ = 'LGPL v2 or later'
__copyright__ = "2013-2015, " + __author__
__url__ = "http://pypi.python.org/pypi/pyditz"

__classifiers__ = """
Development Status :: 4 - Beta
Environment :: Console
Intended Audience :: Developers
License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)
Natural Language :: English
Operating System :: Microsoft :: Windows
Operating System :: POSIX :: Linux
Operating System :: Unix
Programming Language :: Python :: 2.7
Programming Language :: Python :: 3
Programming Language :: Python :: 3.3
Programming Language :: Python :: 3.4
Topic :: Software Development :: Bug Tracking
"""                                     # noqa



"""
Terminal utilities.

http://stackoverflow.com/questions/566746/how-to-get-console-window-width-in-python
"""

import os
import shlex
import struct
import platform
import subprocess


def get_terminal_size(default=(80, 25)):
    """
    Return width and height of console window.
    """

    current_os = platform.system()
    size = None

    if current_os == 'Windows':
        size = _get_terminal_size_windows()
        if size is None:
            size = _get_terminal_size_tput()
            # needed for Windows python in Cygwin's xterm!

    if current_os in ['Linux', 'Darwin'] or current_os.startswith('CYGWIN'):
        size = _get_terminal_size_linux()

    return size or default


def _get_terminal_size_windows():
    try:
        from ctypes import windll, create_string_buffer

        # stdin handle is -10
        # stdout handle is -11
        # stderr handle is -12
        h = windll.kernel32.GetStdHandle(-12)
        csbi = create_string_buffer(22)
        res = windll.kernel32.GetConsoleScreenBufferInfo(h, csbi)

        if res:
            (bufx, bufy, curx, cury, wattr,
             left, top, right, bottom,
             maxx, maxy) = struct.unpack("hhhhHhhhhhh", csbi.raw)
            cols = right - left + 1
            rows = bottom - top + 1
            return cols, rows

    except Exception:
        pass


def _get_terminal_size_tput():
    try:
        cols = int(subprocess.check_call(shlex.split('tput cols')))
        rows = int(subprocess.check_call(shlex.split('tput lines')))
        return cols, rows
    except Exception:
        pass


def _get_terminal_size_linux():
    def ioctl_GWINSZ(fd):
        try:
            import fcntl
            import termios
            data = fcntl.ioctl(fd, termios.TIOCGWINSZ, '1234')
            return struct.unpack('hh', data)
        except Exception:
            pass

    size = ioctl_GWINSZ(0) or ioctl_GWINSZ(1) or ioctl_GWINSZ(2)
    if not size:
        try:
            fd = os.open(os.ctermid(), os.O_RDONLY)
            size = ioctl_GWINSZ(fd)
            os.close(fd)
        except Exception:
            pass

    if not size:
        try:
            size = (os.environ['LINES'], os.environ['COLUMNS'])
        except Exception:
            return None

    return int(size[1]), int(size[0])


if __name__ == "__main__":
    sizex, sizey = get_terminal_size()
    print('width =', sizex, 'height =', sizey)



"""
Test utility functions.
"""

from ditz import util
from ditz import term


def test_utils():
    "Test utility functions"

    for num in 0.43, 1, 23, 64, 432, 3656, 1539424, 41749360, 97492847:
        print(num, '=', util.timespan(num))

    print()
    print(util.extract_username('Alan Partridge <alan@norwichradio.co.uk>'))

    print()
    names = "fred brian derek tarquin ermintrude cyril"
    util.print_columns(names.split(), linelen=45)

    print()
    print(util.default_name())
    print(util.default_email())

    print(util.hostname())
    print(util.editor())


def test_term():
    "Test terminal functions."

    print(term.get_terminal_size())


if __name__ == "__main__":
    test_utils()
    test_term()



"""
Ditz database interface.
"""

import os
import re
import glob

from collections import Counter, defaultdict
from datetime import datetime

from six.moves import cPickle as pickle

from .objects import Config, Project, Issue, Component, Release, find_config
from .files import write_file, write_config, read_object_file
from .util import default_name, default_email, DitzError
from .config import config as userconfig
from .settings import Settings
from .plugin import loader

from .vcs import VCS
from .logger import log

from .flags import (RELEASED, UNRELEASED, BUGFIX, UNSTARTED, STATUS,
                    DISPOSITION)

#: Regexp matching an issue name.
re_issue_name = r'\b([A-Za-z]+-[0-9]+)\b'

#: Regexp matching an issue ID.
re_issue_id = r'{issue ([0-9a-f]+)}'


class DitzDB(object):
    """
    A Ditz issue database.
    """

    project = None
    issues = []

    def __init__(self, project, username=None, email=None, issuedir=None,
                 path=".", settings=None, **kw):
        #: Global settings.
        self.settings = settings or Settings(**kw)

        #: Configuration data.
        self.config = Config(username or default_name(),
                             email or default_email(),
                             issuedir or ".ditz-issues")

        #: Location.
        self.path = os.path.realpath(path)

        #: Project (a :class:`Project` object).
        self.project = Project(project)

        #: List of issues.  Each entry is an :class:`Issue`.
        self.issues = []

        #: Version control handle.
        self.vcs = VCS(self.path)

        #: Mapping of issue IDs to assigned names.
        self._id2name = {}

        #: Mapping of assigned names to issues.
        self._name2issue = {}

    # Database input/output methods.

    @staticmethod
    def read(dirname, settings=None, **kw):
        """
        Read a Ditz database.
        """

        # Create an empty database.
        db = DitzDB("empty", settings=settings, **kw)

        # Find the Ditz config file.
        dirname, config = find_config(dirname, error=True,
                                      search=db.settings.searchparents)
        dirname = os.path.realpath(dirname)

        # Extract the issue directory name.
        issuedir = config.issue_dir

        # Read the project file.
        path = os.path.join(dirname, issuedir, Project.filename)
        log.info("reading project from %s" % path)
        project = read_object_file(path)

        # Read the issues.
        issues = []
        path = os.path.join(dirname, issuedir)
        match = os.path.join(path, Issue.template % "*")
        issue_files = set(glob.glob(match))
        log.info("found %d issues in %s", len(issue_files), path)

        path = os.path.join(dirname, db.settings.cachefile)
        cached_issues = db.readcache(path)
        cache_changed = False

        if cached_issues:
            # Check for new, modified or deleted issues.
            cached_files = set()
            modtime = os.stat(path).st_mtime

            for issue in cached_issues:
                path = os.path.join(dirname, issuedir, issue.filename)

                # Skip issue if it's deleted.
                if not os.path.exists(path):
                    log.info("skipping deleted issue in %s", path)
                    continue

                # Reread issue file if invalid, or newer than cache.
                reread = os.stat(path).st_mtime > modtime

                try:
                    issue.validate()
                except DitzError:
                    reread = True

                if reread:
                    log.info("reading changed issue from %s" % path)
                    issue = read_object_file(path)
                    cache_changed = True

                cached_files.add(path)
                issues.append(issue)

            # Read all the new issues.
            for path in issue_files - cached_files:
                log.info("reading new issue from %s" % path)
                issue = read_object_file(path)
                issues.append(issue)
                cache_changed = True
        else:
            path = os.path.join(dirname, issuedir)
            log.info("reading issues from %s" % path)
            for path in issue_files:
                issue = read_object_file(path)
                issues.append(issue)
                cache_changed = True

        # Initialise database.
        db.path = dirname
        db.config = config
        db.project = project
        db.issues = issues

        if cache_changed:
            db.update_cache()

        db._reassign_names()
        db.vcs = VCS(db.path)

        # Read extra config settings if they exist.
        path = os.path.join(config.issue_dir, "project.cfg")
        if os.path.exists(path):
            log.info("reading %s", path)
            userconfig.read(path)

        # Load any extra plugins from the issue directory.
        path = os.path.join(config.issue_dir, "plugins")
        loader.add_path(path)
        loader.load()

        return db

    def readcache(self, path):
        """
        Read cached issues if possible.
        """

        # Do nothing unless cache is there and wanted.
        if not self.settings.usecache or not os.path.exists(path):
            return None

        # Read cache, checking for errors.
        with open(path, "rb") as fp:
            try:
                issues = pickle.load(fp)
                log.info("found %d issues cached in %s", len(issues), path)
            except Exception as msg:
                log.warning("can't read issue cache (%s)", str(msg))
                issues = None

        # If no issues could be read, remove it.
        if issues is None:
            log.warning("removing unusable issue cache")
            os.unlink(path)

        return issues

    def validate(self):
        """
        Validate the database.  Return the changes made.
        """

        # Changes made.
        changes = defaultdict(list)

        # Update each issue.
        for issue in self.issues:
            changed = False

            # Check issue IDs.
            desc = self.convert_to_id(issue.desc)
            if desc != issue.desc:
                issue.desc = desc
                changes[issue.id].append("fixed issue IDs in description text")
                changed = True

            for idx, event in enumerate(issue.log_events):
                comment = self.convert_to_id(event[3])
                if comment != event[3]:
                    event[3] = comment
                    changes[issue.id].append("fixed issue IDs in "
                                             "log comment %d" % idx)
                    changed = True

            # Check issue UUID.
            uuid = issue.id.lower()
            if uuid != issue.id:
                issue.id = uuid
                changes[issue.id].append("fixed UUID")
                changed = True

            # If issue was changed, write it.
            if changed:
                issue.write(self.issuedir)
                log.info("fixed issue %s" % issue.id)

        # Update issue cache if required.
        if changes:
            self.update_cache()

        return changes

    def write(self, dirname=None):
        """
        Write the database.
        """

        if not dirname:
            dirname = self.path
        else:
            self.path = dirname

        # Write the config file.
        write_config(self.config, dirname)

        # Write the project file.
        path = os.path.join(dirname, self.config.issue_dir)

        projpath = os.path.join(path, self.project.filename)
        exists = os.path.exists(projpath)

        self.project.write(path)
        if self.vcs and self.settings.versioncontrol and not exists:
            self.vcs.add(projpath)

        # Write the issues.
        for issue in self.issues:
            issue.write(path)

        # Update issue cache.
        self.update_cache()

    def save(self, issue=None):
        """
        Save database file if required.
        """

        if not self.settings.autosave:
            return

        if issue:
            log.info("saving issue %s" % self.issue_name(issue))
            path = issue.write(self.issuedir)
            self.update_cache()
        else:
            log.info("saving project")
            path = self.project.write(self.issuedir)

        return path

    def update_cache(self):
        """
        Update the issue cache.
        """

        if self.settings.usecache:
            path = os.path.join(self.path, self.settings.cachefile)
            with open(path, "wb") as fp:
                pickle.dump(self.issues, fp)

            log.info("wrote %d cached issues to %s", len(self.issues), path)

    # High-level database commands.

    def add_issue(self, title, desc="", type=BUGFIX, status=UNSTARTED,
                  disposition=None, creation_time=None, reporter=None,
                  component=None, release=None, comment=None):
        """
        Add a new issue.
        """

        issue = Issue(title, self.convert_to_id(desc), type, status,
                      disposition, creation_time,
                      reporter or self.config.username)

        self.set_component(issue, component, event=False)
        self.set_release(issue, release, event=False)

        self.issues.append(issue)
        self.add_event(issue, "created", comment)
        self._reassign_names()
        self.save(issue)

        if self.vcs and self.settings.versioncontrol:
            path = self.issue_filename(issue)
            self.vcs.add(path)

        return issue

    def add_component(self, name):
        """
        Add a new component.
        """

        comp = Component(name)

        if name not in self.components:
            self.project.components.append(comp)
        else:
            raise DitzError("component already exists: %s" % name)

        self.save()
        return comp

    def add_release(self, name, status=UNRELEASED, release_time=None,
                    comment=None):
        """
        Add a new release.
        """

        if not self.get_release(name):
            rel = Release(name, status, release_time)
            self.add_event(rel, "created", comment)
            self.project.releases.append(rel)
        else:
            raise DitzError("release already exists: %s" % name)

        self.save()
        return rel

    def add_reference(self, issue, reference, comment=None):
        """
        Add a reference to an issue.
        """

        num = issue.add_reference(reference)
        self.add_event(issue, "added reference %d" % num, comment)
        self.save(issue)

    def set_status(self, issue, status, disposition=None, comment=None):
        """
        Set the status of an issue.
        """

        prevstatus = issue.status

        issue.set_status(status)
        issue.set_disposition(disposition)

        if disposition:
            text = "closed with disposition %s" % DISPOSITION[disposition]
        else:
            text = "changed status from %s to %s" % \
                   (STATUS[prevstatus], STATUS[status])

        self.add_event(issue, text, comment)
        self.save(issue)

    def set_component(self, issue, component=None, comment=None, event=True):
        """
        Set the component of an issue.
        """

        if not component:
            component = self.project.name
        elif component not in self.components:
            raise DitzError("unknown component: %s" % component)

        if event:
            text = "assigned to component %s from %s" \
                   % (component, issue.component)
            self.add_event(issue, text, comment)

        issue.component = component
        self._reassign_names()
        self.save(issue)

    def set_release(self, issue, release=None, comment=None, event=True):
        """
        Set the release of an issue.
        """

        if release and not self.get_release(release):
            raise DitzError("unknown release: %s" % release)

        if issue.release == release:
            return

        if event:
            if release:
                text = "assigned to release " + release
                if issue.release:
                    text += " from release " + issue.release
                else:
                    text += " from unassigned"
            else:
                text = "unassigned from release %s" % issue.release

            self.add_event(issue, text, comment)

        issue.release = release
        self.save(issue)

    def release_release(self, name, comment=None):
        """
        Release a release.
        """

        rel = self.get_release(name)
        if not rel:
            raise DitzError("unknown release: %s" % name)

        if rel.released:
            raise DitzError("release '%s' is already released" % name)

        count = 0
        for issue in self.issues:
            if issue.release == name:
                count += 1
                if not issue.closed:
                    raise DitzError("open issue %s must be reassigned"
                                    % self.issue_name(issue))

        if count == 0:
            raise DitzError("no issues assigned to release '%s'" % name)

        rel.status = RELEASED
        rel.release_time = datetime.now()

        self.add_event(rel, "released", comment)
        self.save()

    def archive_release(self, name, path):
        """
        Archive a release.
        """

        rel = self.get_release(name)
        if not rel:
            raise DitzError("unknown release: %s" % name)

        if not rel.released:
            raise DitzError("release '%s' has not been released" % name)

        keep = []
        archive = []
        for issue in self.issues:
            if issue.release == name:
                archive.append(issue)
            else:
                keep.append(issue)

        if not archive:
            raise DitzError("no issues assigned to release %s" % name)

        try:
            os.makedirs(path)
        except OSError as msg:
            raise DitzError("can't create %s: %s" % (path, str(msg)))

        self.project.write(path)
        for issue in archive:
            issue.write(path)

        for issue in archive:
            self.remove_issue(issue)

        self.issues = keep
        self._reassign_names()

        self.save()

    def edit_issue(self, issue, issuetext):
        """
        Edit an issue, by replacing its text and rereading it.
        """

        # Remove the issue.
        self.issues.remove(issue)

        # Write new issue text.
        path = os.path.join(self.issuedir, issue.filename)
        write_file(path, issuetext)

        # Reread the issue.
        issue = read_object_file(path)
        self.issues.append(issue)
        self._reassign_names()

        self.save()

    def drop_issue(self, issue):
        """
        Drop an issue.
        """

        self.remove_issue(issue)
        self.issues.remove(issue)
        self._reassign_names()

        self.save()

    def remove_issue(self, issue):
        """
        Remove an issue file.
        """

        path = self.issue_filename(issue)

        if os.path.exists(path):
            os.remove(path)

        if self.vcs and self.settings.versioncontrol:
            self.vcs.remove(path)

    def add_comment(self, issue, comment):
        """
        Add a comment to an issue.
        """

        self.add_event(issue, "commented", comment)
        self.save(issue)

    def export(self, fmt, path):
        """
        Export issue database to the specified path.
        """

        from .exporter import get_exporter

        exporter = get_exporter(fmt)
        if not exporter:
            raise DitzError("unknown export format: %s" % fmt)

        exporter(self).export(path)

    # Low-level commands and properties.

    def add_event(self, item, text, comment=None):
        """
        Add an event to a database item.
        """

        if comment:
            comment = self.convert_to_id(comment)

        item.event(self.config.username, text, comment)

    def get_issue(self, name):
        """
        Return an issue given its assigned issue name.
        """

        return self._name2issue.get(name, None)

    def issue_name(self, issue):
        """
        Return the assigned issue name of an issue.
        """

        return self._id2name.get(issue.id, "<invalid-issue>")

    def issue_filename(self, issue):
        """
        Return the filename of an issue.
        """

        return os.path.join(self.path, self.config.issue_dir, issue.filename)

    def get_release(self, name):
        """
        Return a release given its name.
        """

        name = str(name)
        for rel in self.project.releases:
            if rel.name == name:
                return rel

        return None

    def get_releases(self, name=None):
        """
        Return list of all releases, or a named release.
        """

        if not name:
            releases = self.project.releases
            return [r for r in releases if not r.released] + [None]

        return [self.get_release(name)]

    @property
    def issue_names(self):
        """
        All existing issue names.
        """

        return sorted(self._name2issue.keys())

    @property
    def issue_events(self):
        """
        Yield all issue-related events and their issues.

        Each yielded event is a tuple containing:

        - date (:class:`datetime`)
        - user (string)
        - description (string)
        - comment (string)
        - issue (:class:`Issue`)
        """

        for issue in self.issues:
            for date, user, text, comment in issue.log_events:
                yield date, user, text, comment, issue

    @property
    def components(self):
        """
        List of defined components.
        """

        return [comp.name for comp in self.project.components]

    @property
    def releases(self):
        """
        List of defined releases.
        """

        return [rel.name for rel in self.project.releases]

    # Issue name/ID handling methods.

    def convert_to_id(self, text):
        """
        Replace names with ``{issue ...}`` in text.

        Args:
            text (str): Text to convert.

        Returns:
            Converted text (str).
        """

        def repl(m):
            issue = self._name2issue.get(m.group(1), None)

            if issue:
                return "{issue %s}" % issue.id
            else:
                return m.group(0)

        return re.sub(re_issue_name, repl, text)

    def convert_to_name(self, text, idmap={}):
        """
        Replace ``{issue ...}`` with names in issue text.

        Args:
            text (str): Text to convert.
            idmap (dict, optional): ID mapping.

        If the ID mapping is specified and contains the issue's ID as a
        key, the replacement is the mapping value.  Otherwise, it's the
        issue's name.

        Returns:
            Converted text (str).
        """

        def repl(m):
            idx = m.group(1)
            if idx in idmap:
                return idmap[idx]
            elif idx in self._id2name:
                return self._id2name[idx]
            else:
                return "[unknown issue %s]" % idx

        return re.sub(re_issue_id, repl, text)

    def related_issues(self, issue):
        """
        Return set of all issues related to an issue.

        That is, those that mention this one, and those that this one
        mentions.
        """

        return self.issues_mentioning(issue) | self.mentioned_issues(issue)

    def relation_mapping(self):
        """
        Return a mapping of each issue to the set of its related issues.
        """

        mapping = defaultdict(set)

        for issue in self.issues:
            for other in self.mentioned_issues(issue):
                mapping[issue].add(other)
                mapping[other].add(issue)

        return mapping

    def issues_mentioning(self, issue):
        """
        Return set of all issues mentioning an issue.
        """

        return set([i for i in self.issues
                    if issue in self.mentioned_issues(i)])

    def mentioned_issues(self, issue):
        """
        Return set of all issues mentioned in an issue's text.
        """

        issues = set()
        for ref in self._issues_in_text(issue.desc):
            issues.add(ref)

        for event in issue.log_events:
            for ref in self._issues_in_text(event[3]):
                issues.add(ref)

        return issues

    def _issues_in_text(self, text):
        """
        Yield all the issues found in the given text.
        """

        for match in re.finditer(re_issue_id, text):
            idx = match.group(1)
            if idx in self._id2name:
                name = self._id2name[idx]
                yield self._name2issue[name]

    def _reassign_names(self):
        """
        Reassign issue names after changes to issue status.
        """

        self.issues.sort()
        counts = Counter()

        self._id2name = {}
        self._name2issue = {}

        for issue in sorted(self.issues, key=lambda x: x.creation_time):
            comp = issue.component.lower()
            counts[comp] += 1
            name = "%s-%d" % (comp, counts[comp])
            self._id2name[issue.id] = name
            self._name2issue[name] = issue

    # Miscellaneous other stuff.

    @property
    def issuedir(self):
        return os.path.join(self.path, self.config.issue_dir)

    def __iter__(self):
        return iter(self.issues)

    def __repr__(self):
        return "<DitzDB: %s>" % self.path



"""
Main console program.
"""

from __future__ import print_function

import os
import sys
import argparse

from .settings import Settings
from .config import config, config_path
from .logger import init_logging
from .plugin import loader
from .util import terminal_size
from .version import __version__

# Name of the main program.
progname = "pyditz"


def main(args=sys.argv[1:]):
    # Build command parser.
    parser = argparse.ArgumentParser(prog=progname)

    parser.add_argument("--version", action="version",
                        version="%(prog)s version " + __version__)

    parser.add_argument("-v", "--verbose", action="store_true",
                        help="be verbose about things")

    parser.add_argument("--traceback", action="store_true",
                        help="print traceback on error")

    group = parser.add_argument_group("global arguments")

    group.add_argument("-i", "--issue-dir", type=str, metavar="DIR",
                       help="use the given issue directory")

    group.add_argument("-u", "--user-config", type=str, metavar="FILE",
                       help="use the given user config file")

    group.add_argument("-c", "--config", type=str, action="append",
                       metavar="SECTION.OPTION=VALUE",
                       help="set a config option explicitly")

    group.add_argument("-S", "--no-search", action="store_true",
                       help="turn off searching of parent directories")

    group.add_argument("-P", "--no-pager", action="store_true",
                       help="turn off paging of output")

    group.add_argument("-H", "--no-highlight", action="store_true",
                       help="turn off syntax highlighting")

    group.add_argument("-X", "--no-plugins", action="store_true",
                       help="turn off loading of external plugins")

    group.add_argument("-V", "--no-vcs", action="store_true",
                       help="turn off use of version control")

    group = parser.add_argument_group("command arguments")

    group.add_argument("-m", "--comment", type=str, metavar="STRING",
                       help="specify a comment")

    group.add_argument("-n", "--no-comment", action="store_true",
                       help="skip asking for a comment")

    # Prepend arguments from environment, if any.
    flags = os.environ.get("DITZFLAGS", None)
    if flags:
        args = flags.split() + args

    # Parse arguments.
    opts, args = parser.parse_known_args(args)

    try:
        # Initialise.
        init_logging(opts.verbose)

        if opts.user_config:
            config.set_file(opts.user_config)

        if opts.config:
            for setting in opts.config:
                config.set_option(setting)

        settings = Settings(autosave=True, usecache=True)

        if opts.no_highlight:
            settings.highlight = False
        else:
            settings.highlight = config.getboolean('highlight', 'enable')

        if opts.no_vcs:
            settings.versioncontrol = False
        else:
            settings.versioncontrol = config.getboolean('vcs', 'enable')

        settings.externalplugins = not opts.no_plugins
        settings.searchparents = not opts.no_search
        settings.termlines = 0 if opts.no_pager else terminal_size()[1]
        settings.nocomment = opts.no_comment
        settings.comment = opts.comment

        # Set up plugin load paths.
        if settings.externalplugins:
            # Load plugins from user config directory.
            path = config_path("plugins")
            loader.add_path(path)

            # Load setuptools plugins.
            loader.add_entrypoint('ditz.plugin')

        # Load plugins.
        loader.load()

        # Build common options to pass to Ditz command.
        path = opts.issue_dir or "."

        # Run things.
        from .commands import DitzCmd

        if not args:
            cmd = DitzCmd(path, settings=settings, interactive=True)
            cmd.cmdloop()
        else:
            cmd = DitzCmd(path, settings=settings)
            command = " ".join(args)
            if not cmd.onecmd(command):
                sys.exit(99)

    except KeyboardInterrupt:
        if opts.traceback:
            raise
        else:
            print()
            sys.exit("%s: aborted" % progname)

    except Exception as msg:
        if opts.traceback:
            raise
        else:
            sys.exit("%s: error: %s" % (progname, msg))


if __name__ == "__main__":
    main()



"""
Python implementation of Ditz (http://rubygems.org/gems/ditz).
"""

import pkg_resources as pkg

from .plugin import loader

path = pkg.resource_filename(__name__, "plugins")
loader.add_path(path)

del path, loader, pkg



"""
Command tests.
"""

import os
import shutil

import ditz.util
from ditz.database import DitzDB
from ditz.commands import DitzCmd
from ditz.files import read_yaml_file

from mock import Mock


def test_commands(name="example"):
    "Test commands"

    ditz.util.editor = Mock(return_value='nosuch')
    ditz.util.run_editor = Mock()

    dbname = name + "-cmd"
    if os.path.isdir(dbname):
        shutil.rmtree(dbname)

    db = DitzDB.read(name)
    db.write(dbname)

    for test in read_yaml_file("test-cmd.yaml"):
        cmd = test.pop("cmd")
        for testcase in test.pop("tests"):
            args = testcase.pop("args") or ""
            command = cmd + " " + str(args)
            print(("=== Running '%s' ===" % command))
            run_command(dbname, command, **testcase)


def run_command(dbname, cmd, **kw):
    ditzcmd = DitzCmd(dbname, usecache=True, autosave=True, highlight=False)

    for method in "line", "text", "choice", "yesno":
        if method in kw:
            effect = kw[method]
        else:
            effect = Exception('no %s input' % method)

        setattr(ditzcmd, "get" + method, Mock(spec=DitzCmd,
                                              side_effect=effect))

    ditzcmd.onecmd(cmd)


if __name__ == "__main__":
    test_commands()



"""
Output syntax highlighting.
"""

import sys

from .config import config
from .logger import log

try:
    from pygments import highlight
    from pygments.lexer import RegexLexer, bygroups
    from pygments.token import Text, Generic
    from pygments.styles import STYLE_MAP

    from pygments.formatters import (NullFormatter,
                                     TerminalFormatter,
                                     Terminal256Formatter)

    class DitzLexer(RegexLexer):
        """
        Interactive Ditz lexer.
        """

        name = "Ditz"
        aliases = ['ditz', 'pyditz']

        tokens = {
            'root': [
                # Ditz prompt.
                (r'^Ditz:', Generic.Prompt),

                # Pager prompt.
                (r'^-- More .+$', Generic.Prompt),

                # Issue tag.
                (r'[a-z]+-\d+', Generic.Strong),

                # Feature indicator.
                (r'(?<=\()feature(?=\))', Generic.Inserted),

                # Bug indicators.
                (r'(?<=\()bug(?=\))', Generic.Deleted),
                (r'(?<=\* )bugfix(?=:)', Generic.Deleted),

                # Release lines.
                (r'^\d[^ \n]+(?= \()', Generic.Heading),
                (r'^Unassigned', Generic.Heading),

                # Closure indicators.
                (r'(?<=closed: )fixed', Generic.Inserted),
                (r'(?<=closed: )won\'t fix', Generic.Deleted),
                (r'(?<=closed: )reorganized', Generic.Strong),

                # Issue subheading.
                (r'^[A-Za-z ]{11}:', Generic.Subheading),

                # Email address.
                (r'[\w.]+@[\w.]+', Generic.Strong),

                # A 'changelog' title.
                (r'^(== +)([^ ]+)( +/ +)(.+)$',
                 bygroups(Text, Generic.Heading,
                          Text, Generic.Subheading)),

                # Error message.
                (r'^Error:', Generic.Error),

                # Everything else.
                (r'.', Text)
            ],
        }

    class DitzSessionLexer(RegexLexer):
        """
        Ditz output lexer.
        """

        name = "Ditz session"
        aliases = ['ditzsession']

        tokens = {
            'root': [
                # Special cases.
                (r'[_=>x] +[a-z]+-\d+: .+$', Text),
                (r'^( +Title|Description):.*$', Text),
                (r'^(Start|Stopp|Comment|Add|Chang|Clos)ing.+$', Text),

                # Prompt and response.
                (r'^([A-Z][^:?\n]*[:?])(.*)$',
                 bygroups(Text, Generic.Strong)),

                # User-typed comment.
                (r'^(>)(.*)$',
                 bygroups(Text, Generic.Strong)),

                # Everything else.
                (r'.*\n', Text)
            ],
        }

    # Set up the terminal formatter.
    lexer = DitzLexer(encoding='utf-8', ensurenl=False)

    if sys.platform == 'win32':
        try:
            import colorama
            colorama.init()
            formatter = TerminalFormatter()
        except ImportError:
            formatter = NullFormatter()
    else:
        style = config.get('highlight', 'style')

        if style in STYLE_MAP:
            formatter = Terminal256Formatter()
        else:
            log.warning("'%s' is not a Pygments style" % style)
            formatter = NullFormatter()

    if not sys.stdout.isatty():
        formatter = NullFormatter()

    def colorize(text):
        return highlight(text, lexer, formatter)
except ImportError:                     # pragma: no cover
    def colorize(text):
        return text



"""
Interactive session recording.
"""

from __future__ import print_function

import re
import sys
import fileinput
from six import StringIO

try:
    import pexpect
except ImportError:
    print("You need to install the pexpect module to generate session")
    print("output.  This also implies using a system that supports it")
    print("(i.e., not Windows).")
    sys.exit(99)


class CmdSession(object):
    def __init__(self, program, prompt):
        self.child = pexpect.spawn(program, timeout=1)
        self.atprompt = False
        self.prompt = prompt
        self.flush()

    def command(self, prompt, response):
        if prompt is None:
            if not self.showing(self.prompt):
                self.output(self.prompt)
            self.toprompt()
        else:
            self.child.expect(prompt)

        self.child.sendline(response)
        self.atprompt = False

    def showing(self, text):
        return self.getoutput()[-len(text):] == text

    def toprompt(self):
        if not self.atprompt:
            self.child.expect(self.prompt)
            self.atprompt = True

    def getoutput(self):
        return self.child.logfile_read.getvalue()

    def output(self, text):
        self.child.logfile_read.write(text)

    def flush(self):
        self.child.logfile_read = StringIO()

    def quit(self, cmd='quit'):
        self.toprompt()
        self.child.sendline(cmd)
        self.child.expect(pexpect.EOF)


def run_session(debug=False):
    command = """
    pyditz --no-search --no-pager --no-plugins --no-highlight --no-vcs
    -c ui.name=Dilbert
    -c ui.email=dilbert@cubicle.com
    """.strip().replace("\n", ' ')

    ditz = CmdSession(command, "Ditz: ")

    for line in fileinput.input():
        m = re.match(r'\.\. command: *(.*)', line)
        if m:
            cmd = m.group(1)
            ditz.command(None, cmd)
            continue

        m = re.match(r'\.\. prompt: *(.*)', line)
        if m:
            prompt = m.group(1) or None
            continue

        m = re.match(r'\.\. reply: *(.*)', line)
        if m:
            reply = m.group(1)
            ditz.command(prompt, reply)
            continue

        m = re.match(r'\.\. (literal)?include:: */(.+\.txt)', line)
        if m:
            recfile = m.group(2)

            ditz.toprompt()
            output = ditz.getoutput()
            lines = output.split("\r\n")[:-1]
            output = "\n".join(lines).rstrip() + "\n"
            ditz.flush()

            with open(recfile, "w") as fp:
                fp.write(output)

            if debug:
                print()
                print("===", recfile, "===")
                print()
                print(output)
            else:
                print("Wrote", recfile)

    ditz.quit()


if __name__ == "__main__":
    run_session()



"""
File utilities.
"""

import sys
import yaml
import codecs
import six

from datetime import datetime

from .config import config


def read_file(path):
    "Read a file."

    with codecs.open(path, "rb", encoding='utf-8') as fp:
        return fp.read()


def write_file(path, text):
    "Write a file."

    with codecs.open(path, "wb", encoding='utf-8') as fp:
        fp.write(text)


def read_yaml_file(path):
    "Read YAML data from a file."

    with codecs.open(path, "rb", encoding='utf-8') as fp:
        return read_yaml(fp)


def read_yaml(fp):
    "Read YAML data from a stream."
    return yaml.safe_load(fp)


def read_object_file(path):
    """
    Read a YAML object from a file and validate it.
    """

    obj = read_yaml_file(path)
    obj.validate()

    return obj


def write_yaml_file(path, item):
    """
    Write an object to a YAML file.
    """

    with codecs.open(path, "wb", encoding='utf-8') as fp:
        fp.write("--- ")
        write_yaml(item, fp)


def write_yaml(item, fp=None, parent=None, level=0):
    """
    Write YAML data to a stream.
    """

    # This should be using PyYAML's write function, but I can't get it to
    # write things the way I want -- i.e., reproduce the format that
    # rubyditz uses.  And datetimes don't get written in a format that
    # rubyditz can read back again.  Sigh.

    # This function is ghastly and evil, and shouldn't exist.  But it gets
    # the job of roundtripping done.

    if fp is None:
        fp = sys.stdout

    value = six.text_type(item)
    tag = getattr(item, "yaml_tag", None)
    indent = "  "

    if tag:
        fp.write(tag + " \n")
        seenref = False
        for attr in item.attributes:
            if attr == "id" and not seenref:
                fp.write("\n")

            fp.write(indent * (level - 1))
            fp.write("%s: " % attr)
            obj = getattr(item, attr)

            if obj and attr == "references":
                seenref = True

            write_yaml(obj, fp, item, level + 1)
    elif isinstance(item, dict):
        if parent:
            fp.write("\n")

        for key, obj in sorted(item.items()):
            fp.write(indent * (level - 1))
            fp.write("%s: " % key)
            write_yaml(obj, fp, item, level + 1)
    elif isinstance(item, list):
        if len(item) > 0:
            newline = doindent = not isinstance(parent, list)

            # This is a gross hack.  Look away now.
            if hasattr(parent, "ditz_tag") and level > 1:
                level -= 1

            if newline:
                fp.write("\n")

            for obj in item:
                if doindent:
                    fp.write(indent * (level - 1))
                else:
                    doindent = True

                fp.write("- ")
                write_yaml(obj, fp, item, level + 1)
        else:
            fp.write("[]\n")
    elif isinstance(item, datetime):
        fp.write("%s Z\n" % six.text_type(item))
    elif "\n" in value:
        fp.write("|-\n")

        if isinstance(parent, list) and level > 1:
            level -= 1

        for line in value.split("\n"):
            fp.write(indent * level)
            fp.write(line + "\n")
    elif item is None:
        fp.write("\n")
    else:
        quote = False
        if not value:
            quote = True
        elif value[0] in '"{}[]':
            quote = True
        elif value[0] != ':' and ":" in value:
            quote = True
        elif '#' in value:
            quote = True
        else:
            try:
                float(value)
                quote = True
            except ValueError:
                pass

        if quote:
            value = '"' + value.replace('"', r'\"') + '"'

        fp.write(value + "\n")


def write_config(conf, dirname="."):
    """
    Write Ditz config file if required.
    """

    if config.getboolean('config', 'create_ditz_config'):
        conf.write(dirname)
        return True

    return False



# -*- coding: utf-8 -*-
#
# PyDitz documentation build configuration file
#
# This file is execfile()d with the current directory set to its containing
# dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys, os

# If extensions (or modules to document with autodoc) are in another
# directory, add these directories to sys.path here. If the directory is
# relative to the documentation root, use os.path.abspath to make it
# absolute, like shown here.
for name in '.', '..':
    sys.path.insert(0, os.path.abspath(name))

# -- General configuration -----------------------------------------------------

import ditz.pkginfo as info
from ditz.version import __version__
from ditz.colorize import DitzLexer, DitzSessionLexer

# Setup function.
def setup(app):
    app.add_lexer('ditz', DitzLexer())
    app.add_lexer('ditzsession', DitzSessionLexer())

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ['sphinx.ext.todo',
              'sphinx.ext.autodoc',
              'sphinx.ext.viewcode',
              'sphinx.ext.napoleon']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
#source_encoding = 'utf-8'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = info.__title__
copyright = info.__copyright__

# The version info for the project you're documenting, acts as replacement
# for |version| and |release|, also used in various other places throughout
# the built documents.
#
# The short X.Y version.
version = __version__
# The full version, including alpha/beta/rc tags.
release = version

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#language = None

# There are two options for replacing |today|: either, you set today to
# some non-false value, then it is used:
#today = ''
# Else, today_fmt is used as the format for a strftime call.
#today_fmt = '%B %d, %Y'

# List of documents that shouldn't be included in the build.
unused_docs = []

# List of directories, relative to source directory, that shouldn't be
# searched for source files.
exclude_trees = []

# The reST default role (used for this markup: `text`) to use for all
# documents.
#default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'default'

# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []

# Common substitutions.
rst_epilog = """
.. |PD|     replace:: PyDitz
.. |CMD|    replace:: ``pyditz``
.. |PDP|    replace:: ``Ditz:``
.. |RC|     replace:: ``ditz.cfg``
.. |RCDIR|  replace:: ``.ditz``
.. |RCPATH| replace:: :samp:`{HOMEDIR}/.ditz`
.. |CONF|   replace:: ``.ditz-config``

.. |GRIN|   image:: /images/face-smile.png

.. _Cmd: http://docs.python.org/2.7/library/cmd
.. _Ditz: http://rubygems.org/gems/ditz
.. _Jinja2: http://jinja.pocoo.org
.. _PEP8: http://www.python.org/dev/peps/pep-0008
.. _PyDitz: https://bitbucket.org/zondo/pyditz
.. _PyPI: https://pypi.python.org
.. _PyYAML: https://pypi.python.org/pypi/PyYAML
.. _Tkinter: https://wiki.python.org/moin/TkInter
.. _flake8: https://pypi.python.org/pypi/flake8
.. _setuptools: http://pythonhosted.org/setuptools
.. _sorttable: http://www.kryogenix.org/code/browser/sorttable
"""

# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
html_theme = 'alabaster'

# Theme options are theme-specific and customize the look and feel of a
# theme further.  For a list of options available for each theme, see the
# documentation.
#html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
html_theme_path = ['.']

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
#html_title = None

# A shorter title for the navigation bar.  Default is the same as
# html_title.
#html_short_title = None

# The name of an image file (relative to this directory) to place at the
# top of the sidebar.
#html_logo = "static/logo.png"

# The name of an image file (within the static path) to use as favicon of
# the docs.  This file should be a Windows icon file (.ico) being 16x16 or
# 32x32 pixels large.
#html_favicon = None

# Add any paths that contain custom static files (such as style sheets)
# here, relative to this directory. They are copied after the builtin
# static files, so a file named "default.css" will overwrite the builtin
# "default.css".
html_static_path = ['issues']

# If not '', a 'Last updated on:' timestamp is inserted at every page
# bottom, using the given strftime format.
#html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
#html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
html_use_modindex = True

# If false, no index is generated.
html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = True

# If true, an OpenSearch description file will be output, and all pages
# will contain a <link> tag referring to it.  The value of this option must
# be the base URL from which the finished HTML is served.
#html_use_opensearch = ''

# If nonempty, this is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = ''

# Output file base name for HTML help builder.
htmlhelp_basename = 'gindoc'


# -- Options for LaTeX output --------------------------------------------------

# The paper size ('letter' or 'a4').
#latex_paper_size = 'letter'

# The font size ('10pt', '11pt' or '12pt').
#latex_font_size = '10pt'

# Grouping the document tree into LaTeX files. List of tuples (source start
# file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
  (master_doc, 'pyditz.tex', u'PyDitz Documentation',
   info.__author__, 'manual'),
]

# The name of an image file (relative to this directory) to place at the
# top of the title page.
#latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are
# parts, not chapters.
#latex_use_parts = False

# Additional stuff for the LaTeX preamble.
#latex_preamble = ''

# Documents to append as an appendix to all manuals.
#latex_appendices = []

# If false, no module index is generated.
#latex_use_modindex = True

# -- Options for PDF output --------------------------------------------------

# Grouping the document tree into PDF files. List of tuples (source start
# file, target name, title, author, documentclass [howto/manual]).
pdf_documents = [
  (master_doc, 'pyditz', u'PyDitz Documentation',
   info.__author__, 'manual'),
]

# -- Options for autodoc -----------------------------------------------------

autodoc_member_order = 'bysource'

autodoc_default_flags = ['members']



"""
Issue commands.
"""

from ditz import flags
from ditz.editor import DitzEditor


class CmdIssue(object):
    def do_add(self, arg):
        "add -- Add an issue"

        title = self.getline("Title: ", allowempty=False)
        desc = self.gettext("Description")

        types = {'b': flags.BUGFIX, 'f': flags.FEATURE, 't': flags.TASK}

        while True:
            reply = self.getline("Is this a (b)ugfix, (f)eature or (t)ask? ")
            if reply and reply[0] in types:
                issuetype = types[reply[0]]
                break

        choices = self.db.components
        if len(choices) >= 2:
            comp = self.getchoice("component", choices)
        else:
            comp = None

        release = None
        releases = self.db.project.releases
        choices = [r.name for r in releases if r.status == flags.UNRELEASED]
        if choices and self.getyesno("Assign to a release now?"):
            if len(choices) > 1:
                release = self.getchoice("release", choices)
            else:
                release = choices[0]
                self.write("Assigning to release", release)

        default = self.db.config.username
        prompt = 'Issue creator (enter for "%s"): ' % default
        reporter = self.getline(prompt)
        comment = self.getcomment()

        issue = self.db.add_issue(title, desc, type=issuetype,
                                  reporter=reporter, release=release,
                                  component=comp, comment=comment)

        name = self.last_issuename = self.db.issue_name(issue)
        self.write("Added issue", name)

    def do_comment(self, arg):
        "comment <issue> -- Comment on an issue"

        issue, name = self.getissue(arg, 1)
        self.write("Commenting on %s: %s" % (name, issue.title))

        comment = self.getcomment()

        if comment:
            self.db.add_comment(issue, comment)
            self.write("Comments recorded for", name)
        else:
            self.write("Empty comment, aborted")

    def do_edit(self, arg):
        "edit <issue> -- Edit an issue"

        issue, name = self.getissue(arg, 1)
        self.write("Editing %s: %s" % (name, issue.title))

        path = self.db.issue_filename(issue)
        editor = DitzEditor(path)

        while True:
            text = editor.edit()
            if not editor.error:
                break

            reply = self.getyesno("%s\n\nRe-edit to fix problem?"
                                  % editor.error, True)
            if not reply:
                return

        if editor.modified:
            self.db.edit_issue(issue, text)
            self.write("Updated %s: %s" % (name, issue.title))
        else:
            self.write("No changes made")

    def do_set_component(self, arg):
        "set-component <issue> [component] -- Set an issue's component"

        issue, name = self.getissue(arg, 1)
        comp = self.getarg(arg, 2)

        self.write("Changing the component of issue %s: %s"
                   % (name, issue.title))

        choices = self.db.components
        if len(choices) < 2:
            self.error("this project does not use multiple components")

        if comp and comp not in choices:
            self.error("unknown component: %s" % comp)

        if not comp:
            if issue.component in choices:
                choices.remove(issue.component)

            comp = self.getchoice("component", choices)

        comment = self.getcomment()

        self.db.set_component(issue, comp, comment)
        newname = self.db.issue_name(issue)

        if self.last_issuename == name:
            self.last_issuename = newname

        self.write("Issue %s is now %s.  Other issue names may have changed."
                   % (name, newname))

    def do_add_reference(self, arg):
        "add-reference <issue> -- Add a reference to an issue"

        issue, name = self.getissue(arg, 1)
        self.write("Adding a reference to %s: %s" % (name, issue.title))

        text = self.getline("Reference: ", allowempty=False)
        comment = self.getcomment()

        self.db.add_reference(issue, text, comment=comment)
        self.write("Added reference to", name)

    def do_start(self, arg):
        "start <unstarted_issue> -- Start work on an issue"

        issue, name = self.getissue(arg, 1)

        if issue.status == flags.IN_PROGRESS:
            self.error("already marked as", flags.STATUS[issue.status])

        self.write("Starting work on %s: %s" % (name, issue.title))
        comment = self.getcomment()

        self.db.set_status(issue, flags.IN_PROGRESS, comment=comment)
        self.write("Recorded start of work for", name)

    def do_stop(self, arg):
        "stop <started_issue> -- Stop work on an issue"

        issue, name = self.getissue(arg, 1)

        if issue.status == flags.PAUSED:
            self.error("already marked as", flags.STATUS[issue.status])

        self.write("Stopping work on %s: %s" % (name, issue.title))
        comment = self.getcomment()

        self.db.set_status(issue, flags.PAUSED, comment=comment)
        self.write("Recorded stopping of work for", name)

    def do_assign(self, arg):
        "assign <issue> [release] -- Assign an issue to a release"

        issue, name = self.getissue(arg, 1)

        if issue.release:
            self.write("Issue %s currently assigned to release %s"
                       % (name, issue.release))
        else:
            self.write("Issue %s not currently assigned to any release" % name)

        release = self.getarg(arg, 2)

        if not release:
            releases = self.db.project.releases
            choices = [r.name for r in releases
                       if r.status == flags.UNRELEASED]

            if issue.release and issue.release in choices:
                choices.remove(issue.release)

            if not choices:
                self.error("no other release available")

            if len(choices) > 1:
                release = self.getchoice("release", choices)
            else:
                release = choices[0]

        if release == issue.release:
            self.error("already assigned to release %s" % release)

        if not self.db.get_release(release):
            self.error("no release with name %s" % release)

        self.write("Assigning to release", release)
        comment = self.getcomment()

        self.db.set_release(issue, release, comment=comment)
        self.write("Assigned", name, "to", release)

    def do_unassign(self, arg):
        "unassign <assigned_issue> -- Unassign an issue from any releases"

        issue, name = self.getissue(arg, 1)

        if not issue.release:
            self.error("not assigned to a release")

        relname = issue.release
        self.write("Unassigning %s: %s" % (name, issue.title))
        comment = self.getcomment()

        self.db.set_release(issue, None, comment=comment)
        self.write("Unassigned", name, "from", relname)

    def do_close(self, arg):
        "close <open_issue> -- Close an issue"

        issue, name = self.getissue(arg, 1)

        self.write("Closing issue %s: %s" % (name, issue.title))

        disp = flags.DISPOSITION
        revdisp = {disp[x]: x for x in disp}
        choices = (disp[flags.FIXED], disp[flags.WONTFIX], disp[flags.REORG])
        choice = self.getchoice("disposition", choices)
        disp = revdisp[choice]

        comment = self.getcomment()

        self.db.set_status(issue, flags.CLOSED, disposition=disp,
                           comment=comment)

        self.write("Closed issue", name, "with disposition", choice)

    def do_drop(self, arg):
        "drop <issue> -- Drop an issue"

        issue, name = self.getissue(arg, 1)

        self.db.drop_issue(issue)
        self.write("Dropped %s.  Other issue names may have changed." % name)



"""
Info commands.
"""

import re
import sys
import six

from ditz.config import config as userconfig
from ditz import display


class CmdInfo(object):
    def do_todo(self, arg):
        "todo [-a,all] [release] -- Generate todo list"

        flag = self.getarg(arg, 1)
        if flag in ("-a", "all"):
            release = self.getrelease(arg, 2, optional=True)
            closed = True
        else:
            release = self.getrelease(arg, 1, optional=True)
            closed = False

        text = display.show_todo(self.db, release, closed)
        self.write(text)

    def do_show(self, arg):
        "show <issue> -- Describe a single issue"

        issue, _ = self.getissue(arg, 1)
        text = display.show_issue(self.db, issue)
        self.write(text)

    def do_log(self, arg):
        """log [count] -- Show recent activity

        If no count is given, the value is taken from the 'log_lines'
        configuration value (default: 5).  If a count is given, it becomes
        the new default.  A count of zero means show all log entries."""

        if not hasattr(self, "last_log"):
            self.config.add('lines', 5)
            self.last_log = self.config.getint('lines')

        count = self.getint(arg, 1)
        if count is None:
            count = self.last_log
        else:
            self.last_log = count

        text = display.log_events(self.db, count=count, verbose=True)
        self.write(text)

    def do_shortlog(self, arg):
        """shortlog [count] -- Show recent activity (short form)

        If no count is given, the value is taken from the 'shortlog_lines'
        configuration value (default: 20).  If a count is given, it becomes
        the new default.  A count of zero means show all log entries."""

        if not hasattr(self, "last_shortlog"):
            self.config.add('lines', 20)
            self.last_shortlog = self.config.getint('lines')

        count = self.getint(arg, 1)
        if count is None:
            count = self.last_shortlog
        else:
            self.last_shortlog = count

        text = display.log_events(self.db, count=count)
        self.write(text)

    def do_changelog(self, arg):
        "changelog <release> -- Generate a changelog for a release"

        name = self.getrelease(arg, 1)
        text = display.show_changelog(self.db, name)
        self.write(text)

    def do_list(self, arg):
        "list [<regexp>] -- List all issues, optionally matching a regexp"

        try:
            regexp = re.compile(arg or ".*")
        except re.error as msg:
            self.error("invalid regexp: %s" % six.text_type(msg))

        text = display.show_grep(self.db, regexp)
        self.write(text or "No matching issues")

    def do_status(self, arg):
        "status [release] -- Show project status"

        release = self.getrelease(arg, 1, optional=True)
        text = display.show_status(self.db, release)
        self.write(text)

    def do_releases(self, arg):
        "releases -- Show releases"

        text = display.show_releases(self.db)
        self.write(text or "No releases")

    def do_config(self, arg):
        "config [<section>] -- show configuration settings"

        section = self.getarg(arg, 1)

        if not section:
            userconfig.write(sys.stdout)
        elif userconfig.has_section(section):
            for name in sorted(userconfig.options(section)):
                self.write(name, '=', userconfig.get(section, name))
        else:
            self.error("no config section called '%s'" % section)

    def do_path(self, arg):
        "path -- show the root Ditz database path"

        self.write(self.db.path)



"""
Data validation schemas.
"""

from voluptuous import (Schema, Any, All, MultipleInvalid,
                        ExactSequence, Length, In, Coerce)

from datetime import datetime
from six import text_type

from .flags import TYPE, STATUS, RELSTATUS, DISPOSITION
from .util import DitzError


def validate(obj, schema, filename=None):
    "Check a DitzObject is a valid Ditz object."

    try:
        data = to_dict(obj)
        data = schema(data)
        set_data(obj, data)

    except MultipleInvalid as e:
        if filename:
            raise DitzError("%s: %s" % (filename, str(e)))
        else:
            raise DitzError(str(e))


def to_dict(obj):
    "Convert object to a dict, recursively."

    if hasattr(obj, "__dict__"):
        obj = {attr: to_dict(value) for attr, value in obj.__dict__.items()}
    elif isinstance(obj, list):
        obj = [to_dict(item) for item in obj]

    return obj


def set_data(obj, data):
    "Set object attributes from data, recursively."

    if hasattr(obj, "__dict__"):
        for attr, value in data.items():
            subobj = getattr(obj, attr)
            if not set_data(subobj, value):
                obj.__dict__[attr] = value

        return True

    elif isinstance(obj, list):
        for idx, value in enumerate(data):
            subobj = obj[idx]
            if not set_data(subobj, value):
                obj[idx] = value

        return True

    return False


def Option(options, **kw):
    "Validator for one of a set of options."
    msg = kw.pop('msg', None)
    names = [str(v) for v in options]
    return In(options, msg=msg or "expected one of " + ", ".join(names))


def Optional(type, **kw):
    "Validator for a particular type, or None."
    return Any(type, None, msg=kw.pop('msg', None))


def Text(value):
    "Convert value to text."
    return None if value is None else text_type(value)


# Declare basic Ditz types.
text = Coerce(Text)
issuetype = Option(TYPE.keys())
status = Option(STATUS.keys())
relstatus = Option([None] + list(RELSTATUS.keys()))
disposition = Option([None] + list(DISPOSITION.keys()))
event = All(Length(min=4, max=4, msg="expected a list of length 4"),
            ExactSequence([datetime, text, text, text]))

# Declare schemas for each file object.
issue = Schema({'title': text,
                'desc': text,
                'type': issuetype,
                'component': text,
                'release': Optional(text),
                'reporter': text,
                'status': status,
                'disposition': disposition,
                'creation_time': datetime,
                'references': [text],
                'id': text,
                'log_events': [event]}, required=True)

component = Schema({'name': text}, required=True)

release = Schema({'name': text,
                  'status': relstatus,
                  'release_time': Optional(datetime),
                  'log_events': [event]}, required=True)

project = Schema({'name': text,
                  'version': text,
                  'components': [component],
                  'releases': [release]}, required=True)



"""
Utility functions.
"""

from __future__ import print_function

import os
import re
import sys
import socket
import six

from datetime import datetime

from .config import config
from .term import get_terminal_size
from .logger import log


def age(date, ago=True):
    "Return a human-readable in-the-past time string given a date."
    return timespan((datetime.now() - date).total_seconds(), ago)


def timespan(counter, ago=False):
    "Return approximate timespan for a number of seconds."

    counter = int(counter)
    if counter == 0:
        return "just now"

    for unit, count in (("second", 60),
                        ("minute", 60),
                        ("hour",   24),
                        ("day",     7),
                        ("week",    4),
                        ("month",  12),
                        ("year",    0)):
        if count > 0 and counter >= count * 2:
            counter /= count
        else:
            break

    return "%d %s%s%s" % (counter, unit,
                          "s" if counter > 1 else "",
                          " ago" if ago else "")


def extract_username(email):
    "Return a short user name given email text."

    if '@' not in email:
        return email

    m = re.search('([A-Za-z0-9_.]+)@', email)
    if m:
        return m.group(1)

    return email


def default_name():
    "Return the default user name."

    name = ui_env_value('name', ["DITZUSER", "USER", "USERNAME"])
    if name:
        return name

    return "ditzuser"


def default_email():
    "Return the default email address."

    email = ui_env_value('email', ["DITZEMAIL", "EMAIL"])
    if email:
        return email

    return default_name() + '@' + hostname()


def hostname():
    "Return the host name."

    name = ui_env_value('hostname', ["DITZHOST", "HOSTNAME", "COMPUTERNAME"])
    if name:
        return name

    try:
        return socket.gethostname()
    except socket.error:
        return "UNKNOWN"


def editor():
    "Return a text editor."

    program = ui_env_value('editor', ["DITZEDITOR", "EDITOR", "VISUAL"])
    if program:
        return program

    if sys.platform.startswith('linux'):
        return 'vi'

    return None


def run_editor(program, filename):
    "Run editor program on a file."

    status = os.system('%s "%s"' % (program, filename))
    if status != 0:
        raise DitzError("can't run editor '%s' (status: %d)"
                        % (program, status))


def terminal_size():
    "Return terminal size, or zero if stdout is not a tty."

    if sys.stdout.isatty():
        return get_terminal_size()
    else:
        return (0, 0)


def print_columns(items, linelen=70, spacing=2):
    "Print a number of items in column format."

    maxlen = max(len(text) for text in items)
    columns = max(linelen // (maxlen + spacing), 1)
    padding = " " * spacing

    count = 0
    while count < len(items):
        print(items[count].ljust(maxlen) + padding, end=' ')
        count += 1
        if count % columns == 0:
            print()

    if count % columns != 0:
        print()


def make_directory(path, force=False):
    "Create a directory if it doesn't exist."

    if not os.path.exists(path):
        try:
            os.makedirs(path)
        except OSError as msg:
            raise DitzError(msg)
    elif force:
        raise DitzError("directory '%s' already exists" % path)


def check_value(name, value, choices):
    "Give an error if a value isn't in a set of choices."

    if value not in choices:
        raise ValueError("unknown %s: %s (one of %s expected)"
                         % (name, value, ", ".join(list(choices.keys()))))


def ui_env_value(option, envvars=[]):
    "Return a setting from config [ui] section or environment variables."

    if option:
        value = config.get('ui', option)
        if value:
            return value

    for var in envvars:
        if var in os.environ:
            return os.environ[var]

    return None


def html_markup_function(tag=None):
    "Return a HTML text markup function given a markup tag name."

    def func(text):
        paras = text.strip().split("\n\n")
        paras = ["<p>%s</p>" % p for p in paras]
        return "\n".join(paras)

    if not tag:
        return func

    try:
        import markups

    except ImportError:
        log.info("markups module is not available")

    else:
        cls = markups.find_markup_class_by_name(tag)

        if not cls:
            log.info("markup '%s' is not known", tag)
        elif not cls.available():
            log.info("markup '%s' is not available", tag)
        else:
            func = cls().get_document_body

    return func


def to_unicode(obj, encoding='utf-8'):
    """
    Convert an object to Unicode.

    See http://farmdev.com/talks/unicode
    """

    if six.PY2:
        if isinstance(obj, basestring):
            if not isinstance(obj, unicode):
                obj = unicode(obj, encoding)

    return obj


class DitzError(Exception):
    "A generic Ditz error."



# Setup script for PyDitz.

import os

# Bootstrap setuptools.
from conf.ez_setup import use_setuptools
use_setuptools()

from setuptools import setup

# Set up entry points.
entry_points = """
[console_scripts]
%s = ditz.console:main
""" % os.environ.get("DITZCMD", "pyditz")

# Do the setup.
from conf.tools import read_pkginfo
info = read_pkginfo("ditz")

setup(name             = info.__title__,
      version          = info.__version__, 
      author           = info.__author__,
      author_email     = info.__email__,
      description      = info.__desc__,
      long_description = "\n" + open("README").read(),
      license          = info.__license__,
      url              = info.__url__,
      classifiers      = info.__classifiers__.strip().split("\n"),

      packages = ["ditz"],
      include_package_data = True,
      entry_points = entry_points,

      setup_requires = [
          'hgtools',
          'flake8',
      ],

      install_requires = [
          'pyyaml >= 3.10',
          'jinja2 >= 2.7',
          'voluptuous >= 0.8',
          'six >= 1.8.0',
      ],

      tests_require = [
          'nose >= 1.3.0',
          'mock >= 1.0.1',
          'coverage >= 3.6',
      ],

      test_suite = 'nose.collector')

# flake8: noqa



"""
Display routines.
"""

from itertools import islice
from textwrap import wrap

from . import flags
from . import util


def show_grep(db, regexp):
    """
    Return text describing issues that match a regexp.
    """

    issues = [issue for issue in db.issues if issue.grep(regexp)]

    if issues:
        issues = sorted(issues, reverse=True)
        return "\n".join(list_issues(db, issues))
    else:
        return None


def show_todo(db, name=None, closed=False):
    """
    Return text describing unresolved issues.
    """

    releases = db.get_releases(name)

    lines = []
    add = lines.append

    for num, rel in enumerate(releases):
        if num > 0:
            add('')

        if rel:
            add(rel.description + ":")
        else:
            add("Unassigned:")

        show = []
        for issue in db.issues:
            if rel and issue.release != rel.name:
                continue
            elif not rel and issue.release:
                continue
            elif not closed and issue.closed:
                continue

            show.append(issue)

        if not show:
            add("No issues" if closed else "No open issues")
            continue

        for text in list_issues(db, sorted(show, reverse=True)):
            add(text)

    return "\n".join(lines)


def show_issue(db, issue, underline='-'):
    """
    Return text describing an issue.
    """

    lines = []
    add = lines.append

    def addattr(attr, val=""):
        if val:
            for idx, line in enumerate(wrap(val, width=65)):
                if idx == 0:
                    add("%11s: %s" % (attr, line))
                else:
                    add(" " * 13 + line)
        else:
            add("%11s: " % attr)

    title = "Issue %s" % db.issue_name(issue)
    add(title)
    add(underline * len(title))

    addattr("Title", issue.title)

    desc = db.convert_to_name(issue.desc)

    if "\n" in desc:
        addattr("Description")
        for line in desc.split("\n"):
            add("  " + line)
    else:
        addattr("Description", desc)

    add('')
    addattr("Type", flags.TYPE[issue.type])

    if issue.closed:
        addattr("Status", "closed: " + flags.DISPOSITION[issue.disposition])
    else:
        addattr("Status", flags.STATUS[issue.status])

    addattr("Creator", issue.reporter)
    addattr("Age", util.age(issue.creation_time, False))

    if issue.release:
        addattr("Release", issue.release)

    if issue.references:
        addattr("References")
        for num, ref in enumerate(issue.references, 1):
            add("%3d. %s" % (num, ref))

    addattr("Identifier", issue.id)

    issues = sorted(list(db.related_issues(issue)))
    if issues:
        text = ", ".join([db.issue_name(i) for i in issues])
        addattr("Related", text)

    progress = issue.progresstime
    if progress:
        addattr("In progress", util.timespan(progress))

    add('')
    add("Event log:")

    for date, email, text, comment in reversed(issue.log_events):
        add("- %s (%s, %s)" % (text,
                               util.extract_username(email),
                               util.age(date)))
        if comment:
            for line in db.convert_to_name(comment).split("\n"):
                add('  > ' + line)

    return "\n".join(lines)


def show_releases(db):
    """
    Return text describing all releases.
    """

    lines = [rel.description for rel in reversed(db.project.releases)]
    return "\n".join(lines)


def show_changelog(db, name):
    """
    Return text describing a changelog for a release.
    """

    rel = db.get_release(name)
    if not rel:
        raise util.DitzError("unknown release: %s" % name)

    lines = []
    add = lines.append

    text = "== " + rel.name + " / "
    if rel.released:
        text += rel.release_time.strftime("%Y-%m-%d")
    else:
        text += "unreleased"

    add(text)

    for issue in sorted(db.issues, key=lambda x: x.type):
        if issue.release == rel.name:
            if not issue.closed:
                continue

            if issue.bugfix:
                text = "bugfix: " + issue.title
            else:
                text = issue.title

            add("* " + text)

    return "\n".join(lines)


def show_status(db, name=None, maxflags=20):
    """
    Return text describing the status of a release or releases.
    """

    releases = db.get_releases(name)

    data = []
    for rel in releases:
        reldata = []
        data.append(reldata)
        reldata.append(rel.name if rel else "unassigned")

        issues = []
        alltotal = allclosed = 0
        for itype in flags.BUGFIX, flags.FEATURE, flags.TASK:
            closed = total = 0

            for issue in db.issues:
                if issue.type != itype:
                    continue
                elif rel and issue.release != rel.name:
                    continue
                elif not rel and issue.release:
                    continue

                total += 1
                if issue.closed:
                    closed += 1

                issues.append(issue)

            alltotal += total
            allclosed += closed

            text = "%2d/%2d %s" % (closed, total, flags.TYPE_PLURAL[itype])
            reldata.append(text)

        if not rel:
            text = ""
        elif rel.released:
            text = "(released)"
        elif alltotal == 0:
            text = "(no issues)"
        elif allclosed == alltotal:
            text = "(ready for release)"
        else:
            statusflags = [flags.FLAGS[x.status] for x in sorted(issues)]
            if len(statusflags) > maxflags:
                newflags = []

                for i in range(maxflags):
                    factor = float(i) / (maxflags - 1)
                    idx = int(factor * (len(statusflags) - 1))
                    newflags.append(statusflags[idx])

                statusflags = newflags

            text = "".join(statusflags)

        reldata.append(text)

    maxlen = [0] * 5
    for reldata in data:
        for col in range(5):
            maxlen[col] = max(maxlen[col], len(reldata[col]))

    lines = []
    for reldata in data:
        for col in range(5):
            if col in (0, 4):
                reldata[col] = reldata[col].ljust(maxlen[col])
            else:
                reldata[col] = reldata[col].rjust(maxlen[col])

        lines.append("  ".join(reldata))

    return "\n".join(lines)


def list_issues(db, issues):
    """
    Return text description lines for a list of issues.
    """

    issues = list(issues)
    if not issues:
        return []

    lines = []
    maxlen = max(len(db.issue_name(issue)) for issue in issues)

    for issue in issues:
        text = flags.FLAGS[issue.status] + " "
        text += db.issue_name(issue).rjust(maxlen)
        text += ": " + issue.longname
        lines.append(text)

    return lines


def log_events(db, verbose=False, count=None, datefmt="%a %b %d %X %Y"):
    """
    Return log event message text.
    """

    lines = []
    add = lines.append

    events = sorted(db.issue_events, key=lambda x: x[0], reverse=True)
    shortlog = []
    widths = [0] * 4

    for date, email, text, comment, issue in islice(events, count or None):
        name = db.issue_name(issue)
        when = util.age(date)

        if verbose:
            add('date   : %s (%s)' % (date.strftime(datefmt), when))
            add('author : %s' % email)
            add('issue  : [%s] %s' % (name, issue.title))
            add('')
            add('  ' + text)

            if comment:
                for line in db.convert_to_name(comment).split("\n"):
                    add('  > ' + line)

            add('')
        else:
            user = util.extract_username(email)
            data = [when, name, user, text]
            shortlog.append(data)
            for i in range(4):
                widths[i] = max(widths[i], len(data[i]))

    if not verbose:
        for data in shortlog:
            fields = [data[i].rjust(widths[i]) for i in range(3)]
            fields += [data[3]]
            add(" | ".join(fields))

    return "\n".join(lines)



"""
Internal Ditz objects.
"""

import os
import re
import yaml
import hashlib
import random
import six

from datetime import datetime, timedelta
from functools import total_ordering

from .util import (make_directory, check_value, to_unicode,
                   default_name, default_email, DitzError)

from .flags import (STATUS, RELSTATUS, RELEASED, UNRELEASED, BUGFIX,
                    FEATURE, TASK, DISPOSITION, UNSTARTED, TYPE, SORT,
                    CLOSED)

from .files import read_object_file, write_yaml_file, write_config
from .config import config

from . import schemas

# Regexp matching start of progress on an issue.
re_start = re.compile(r'to in.progress')

# Regexp matching end of progress on an issue.
re_stop = re.compile(r'(from in.progress|closed with disposition)')


class DitzObject(yaml.YAMLObject):
    """
    Base class of a Ditz object appearing in a YAML file.

    Attributes:
        ditz_tag (str): Base YAML tag prefix.
        yaml_tag (str): YAML tag written to file.
        filename (str): YAML filename to read/write.
        attributes (list): List of recognized attributes.
        log_events (list): List of events.
    """

    yaml_loader = yaml.SafeLoader

    ditz_tag = "!ditz.rubyforge.org,2008-03-06"
    yaml_tag = None
    filename = None
    schema = None
    attributes = []

    def write(self, dirname="."):
        """
        Write object to its YAML file.

        Args:
            dirname (str, optional): Directory to write file into.
        """

        make_directory(dirname)
        path = os.path.join(dirname, self.filename)
        write_yaml_file(path, self)

    def event(self, username, text, comment=None, timestamp=None):
        """
        Add an event to the object's ``log_events`` list.

        Args:
            username (str): User name.
            text (str): Event description.
            comment (str, optional): User comment.
            timestamp (datetime, optional): Time of the event.
        """

        time = timestamp or datetime.now()
        text = to_unicode(text)
        comment = to_unicode(comment) or ""
        self.log_events.append([time, username, text, comment])

    def validate(self):
        """
        Check the object is a valid Ditz object.
        """

        if self.schema:
            schemas.validate(self, self.schema, self.filename)
        else:
            raise DitzError("no schema defined for %s"
                            % self.__class__.__name__)


class Project(DitzObject):
    yaml_tag = DitzObject.ditz_tag + '/project'
    filename = "project.yaml"
    schema = schemas.project

    #: Attributes that get saved to file.
    attributes = ["name", "version", "components", "releases"]

    #: Name of the project.
    name = ""

    #: List of components.  Each entry is a :class:`Component`.
    components = []

    #: List of releases.  Each entry is a :class:`Release`.
    releases = []

    def __init__(self, name, version=0.5):
        self.name = to_unicode(name)
        self.version = version
        self.components = [Component(name)]
        self.releases = []


class Release(DitzObject):
    yaml_tag = DitzObject.ditz_tag + '/release'
    schema = schemas.release

    #: Attributes that get saved to file.
    attributes = ["name", "status", "release_time", "log_events"]

    #: Name of the release.
    name = None

    #: Time it was released, or None.
    release_time = None

    #: List of log events.  Each entry is a tuple of (time, username, text,
    #: comment).
    log_events = []

    def __init__(self, name, status=UNRELEASED, release_time=None):
        self.name = to_unicode(name)
        self.status = status
        self.release_time = release_time
        self.log_events = []

        self.set_status(status)

    def set_status(self, status):
        check_value("status", status, RELSTATUS)
        self.status = status

    @property
    def released(self):
        "Whether release has actually been released."
        return self.status == RELEASED

    @property
    def description(self):
        if not self.released:
            tag = "unreleased"
        else:
            tag = "released %s" % (self.release_time.strftime("%Y-%m-%d"))

        return "%s (%s)" % (self.name, tag)

    def __repr__(self):
        return "<Release: %s>" % self.name


class Component(DitzObject):
    yaml_tag = DitzObject.ditz_tag + '/component'
    attributes = ["name"]
    schema = schemas.component

    #: Name of the component.
    name = ""

    def __init__(self, name):
        self.name = to_unicode(name)

    def __repr__(self):
        return "<Component: %s>" % self.name


@total_ordering
class Issue(DitzObject):
    yaml_tag = DitzObject.ditz_tag + '/issue'
    template = "issue-%s.yaml"
    schema = schemas.issue

    #: Attributes that get saved to file.
    attributes = ["title", "desc", "type", "component", "release",
                  "reporter", "status", "disposition", "creation_time",
                  "references", "id", "log_events"]

    #: List of references.  Each entry is a string.
    references = []

    #: List of log events.  Each entry is a tuple of (time, username, text,
    #: comment).
    log_events = []

    def __init__(self, title, desc="", type=BUGFIX, status=UNSTARTED,
                 disposition=None, creation_time=None, reporter=""):
        self.title = to_unicode(title.strip())
        self.desc = to_unicode(desc.strip())
        self.component = None
        self.release = None
        self.reporter = to_unicode(reporter)

        self.set_type(type)
        self.set_status(status)
        self.set_disposition(disposition)

        self.creation_time = creation_time or datetime.now()
        self.id = self.make_id()

        self.references = []
        self.log_events = []

    @property
    def name(self):
        "Name of the issue.  This is the same as its title."
        return self.title

    @property
    def longname(self):
        "Name of the issue, with (bug) or (feature) appended."
        if self.bugfix:
            return self.title + " (bug)"
        elif self.feature:
            return self.title + " (feature)"
        else:
            return self.title

    @property
    def closed(self):
        "Whether the issue is closed."
        return self.status == CLOSED

    @property
    def bugfix(self):
        "Whether the issue is a bugfix."
        return self.type == BUGFIX

    @property
    def feature(self):
        "Whether the issue is a feature."
        return self.type == FEATURE

    @property
    def task(self):
        "Whether the issue is a task."
        return self.type == TASK

    @property
    def filename(self):
        return self.template % self.id

    @property
    def progresstime(self):
        "Total seconds the issue's state has been in-progress."
        total = timedelta()
        start = None

        for time, _, event, _ in self.log_events:
            if re_start.search(event):
                start = time
            elif re_stop.search(event) and start is not None:
                total += time - start
                start = None

        if start is not None:
            total += datetime.now() - start

        return total.total_seconds()

    def set_type(self, type):
        check_value("type", type, TYPE)
        self.type = type

    def set_status(self, status):
        check_value("status", status, STATUS)
        self.status = status

    def set_disposition(self, disposition):
        if disposition:
            check_value("disposition", disposition, DISPOSITION)

        self.disposition = disposition

    def add_reference(self, reference):
        reference = to_unicode(reference)
        self.references.append(reference)
        return len(self.references)

    def grep(self, regexp):
        for attr in "title", "desc", "reporter":
            if regexp.search(getattr(self, attr)):
                return True

        for ref in self.references:
            if regexp.search(ref):
                return True

        for event in self.log_events:
            if regexp.search(event[3]):
                return True

        return False

    def make_id(self):
        data = [self.creation_time, random.random(),
                self.reporter, self.title, self.desc]

        string = "\n".join([six.text_type(item) for item in data])
        return hashlib.sha1(string.encode('utf-8')).hexdigest()

    def __eq__(self, other):
        if hasattr(other, "id"):
            return self.id == other.id

        return False

    def __lt__(self, other):
        this = (SORT[self.status], self.creation_time)
        that = (SORT[other.status], other.creation_time)
        return this < that

    def __hash__(self):
        return int(self.id, 16)

    def __repr__(self):
        return "<Issue: %s>" % self.id


class Config(DitzObject):
    yaml_tag = DitzObject.ditz_tag + '/config'
    filename = ".ditz-config"
    attributes = ["name", "email", "issue_dir"]

    def __init__(self, name, email, issue_dir="issues"):
        self.name = to_unicode(name)
        self.email = to_unicode(email)
        self.issue_dir = issue_dir

    @property
    def username(self):
        return "%s <%s>" % (self.name, self.email)

    def validate(self):
        if not hasattr(self, 'issue_dir'):
            raise DitzError("'issue_dir' not defined")

    def __repr__(self):
        return "<Config: %s>" % self.name


def find_config(basedir=".", error=False, search=True):
    """
    Find Ditz configuration in (or possibly above) a base directory.

    Return the directory and corresponding Config object.
    """

    basedir = os.path.realpath(basedir)
    curdir = basedir

    issuedirs = config.get('config', 'issuedirs').split()

    while True:
        # If Ditz config found, read it.
        path = os.path.join(curdir, Config.filename)
        if os.path.exists(path):
            return curdir, read_object_file(path)

        # If there's an issue directory containing a project file, use
        # that.
        for dirname in issuedirs:
            path = os.path.join(curdir, dirname, Project.filename)
            if os.path.exists(path):
                name = default_name()
                email = default_email()
                conf = Config(name, email, dirname)
                write_config(conf)
                return curdir, conf

        # Otherwise, go to parent directory.
        pardir = os.path.split(curdir)[0]

        # If it's not found, or we're at the top, give up.
        if pardir == curdir or not search:
            if error:
                where = "in or above" if search else "in"
                raise DitzError("can't find '%s' %s '%s'"
                                % (Config.filename, where, basedir))
            else:
                return None, None

        curdir = pardir



"""
VCS support.
"""

import os
from subprocess import check_call, CalledProcessError

from .config import config
from .logger import log


class VCS(object):
    def __init__(self, rootdir):
        self.cmd = None
        self.addcmd = None
        self.removecmd = None

        for name in config.get('vcs', 'systems').split():
            reponame = config.get(name, 'repo')
            repodir = os.path.join(rootdir, reponame)
            if os.path.isdir(repodir):
                log.info("detected '%s' version control system" % name)
                self.cmd = config.get(name, 'cmd')

                if config.has_option(name, 'add'):
                    self.addcmd = config.get(name, 'add')
                else:
                    self.addcmd = 'add'

                if config.has_option(name, 'remove'):
                    self.removecmd = config.get(name, 'remove')
                else:
                    self.removecmd = 'remove'

                break

    def add(self, path):
        self.runcmd(self.addcmd, path)

    def remove(self, path):
        self.runcmd(self.removecmd, path)

    def runcmd(self, subcmd, path):
        if self.cmd and subcmd:
            cmdline = " ".join([self.cmd, subcmd])
            args = cmdline.split() + [path]
            log.info("running '%s'" % " ".join(args))

            try:
                check_call(args)
            except (OSError, CalledProcessError) as err:
                log.info("failed to run '%s': %s" % (args[0], str(err)))
